# This file is part of jax-healpy.
# Copyright (C) 2024 CNRS / SciPol developers
#
# jax-healpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jax-healpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jax-healpy. If not, see <https://www.gnu.org/licenses/>.

from functools import partial, wraps
from typing import Callable, ParamSpec, TypeVar

import jax
import jax.numpy as jnp
from jax.typing import ArrayLike

try:
    from s2fft.recursions.price_mcewen import generate_precomputes_jax
    from s2fft.sampling.reindex import flm_2d_to_hp_fast, flm_hp_to_2d_fast
    from s2fft.transforms import spherical
except ImportError:
    pass

from jax_healpy import npix2nside

__all__ = [
    'alm2map',
    'map2alm',
]

Param = ParamSpec('Param')
ReturnType = TypeVar('ReturnType')


def requires_s2fft(func: Callable[Param, ReturnType]) -> Callable[Param, ReturnType]:
    try:
        import s2fft  # noqa

        return func
    except ImportError:
        pass

    @wraps(func)
    def deferred_func(*args: Param.args, **kwargs: Param.kwargs) -> ReturnType:
        msg = "Missing optional library 's2fft', part of the 'recommended' dependency group."
        raise ImportError(msg)

    return deferred_func


@partial(
    jax.jit,
    static_argnames=[
        'nside',
        'lmax',
        'mmax',
        'pixwin',
        'fwhm',
        'sigma',
        'pol',
        'inplace',
        'verbose',
        'healpy_ordering',
    ],
)
@requires_s2fft
def alm2map(
    alms: ArrayLike,
    nside: int,
    lmax=None,
    mmax=None,
    pixwin=False,
    fwhm=0.0,
    sigma=None,
    pol=True,
    inplace=False,
    verbose=True,
    healpy_ordering: bool = False,
):
    """Computes a Healpix map given the alm.

    The alm are given as a complex array. You can specify lmax
    and mmax, or they will be computed from array size (assuming
    lmax==mmax).

    Parameters
    ----------
    alms : complex, array or sequence of arrays
      A complex array or a sequence of complex arrays.
      Each array must have a size of the form: mmax * (2 * lmax + 1 - mmax) / 2 + lmax + 1
    nside : int, scalar
      The nside of the output map.
    lmax : None or int, scalar, optional
      Explicitly define lmax (needed if mmax!=lmax)
    mmax : None or int, scalar, optional
      Explicitly define mmax (needed if mmax!=lmax)
    pixwin : bool, optional
      Smooth the alm using the pixel window functions. Default: False.
    fwhm : float, scalar, optional
      The fwhm of the Gaussian used to smooth the map (applied on alm)
      [in radians]
    sigma : float, scalar, optional
      The sigma of the Gaussian used to smooth the map (applied on alm)
      [in radians]
    pol : bool, optional
      If True, assumes input alms are TEB. Output will be TQU maps.
      (input must be 1 or 3 alms)
      If False, apply spin 0 harmonic transform to each alm.
      (input can be any number of alms)
      If there is only one input alm, it has no effect. Default: True.
    inplace : bool, optional
      If True, input alms may be modified by pixel window function and beam
      smoothing (if alm(s) are complex128 contiguous arrays).
      Otherwise, input alms are not modified. A copy is made if needed to
      apply beam smoothing or pixel window.
    healpy ordering : bool, optional
      True if the input alms follow the healpy ordering. By default, the s2fft
      ordering is assumed.

    Returns
    -------
    maps : array or list of arrays
      A Healpix map in RING scheme at nside or a list of T,Q,U maps (if
      polarized input)

    Notes
    -----
    Running map2alm then alm2map will not return exactly the same map if the discretized field you construct on the
    sphere is not band-limited (for example, if you have a map containing pixel-based noise rather than beam-smoothed
    noise). If you need a band-limited map, you have to start with random numbers in lm space and transform these via
    alm2map. With such an input, the accuracy of map2alm->alm2map should be quite good, depending on your choices
    of lmax, mmax and nside (for some typical values, see e.g., section 5.1 of https://arxiv.org/pdf/1010.2084).
    """
    if mmax is not None:
        raise NotImplementedError('Specifying mmax is not implemented.')
    if pixwin:
        raise NotImplementedError('Specifying pixwin is not implemented.')
    if fwhm != 0:
        raise NotImplementedError('Specifying fwhm is not implemented.')
    if sigma is not None:
        raise NotImplementedError('Specifying sigma is not implemented.')
    alms = jnp.asarray(alms)
    if alms.ndim == 0:
        raise ValueError('Input alms must have at least one dimension.')
    expected_ndim = 1 if healpy_ordering else 2
    if alms.ndim > expected_ndim + 1 + pol:
        raise ValueError('Input alms have too many dimensions.')
    if alms.ndim == expected_ndim + 1 + pol:
        return jax.vmap(alm2map, in_axes=(0,) + 10 * (None,))(
            alms,
            nside,
            lmax,
            mmax,
            pixwin,
            fwhm,
            sigma,
            pol,
            inplace,
            False,
            healpy_ordering,
        )
    if alms.ndim > expected_ndim:
        # only happens if pol=True
        raise NotImplementedError('TEB alms are not implemented.')

    if lmax is None:
        L = 3 * nside
    else:
        L = lmax + 1

    if healpy_ordering:
        alms = flm_hp_to_2d_fast(alms, L)

    sampling = 'healpix'
    method = 'jax'
    spmd = False

    precomps = generate_precomputes_jax(L, 0, sampling, nside, False)
    f = spherical.inverse(
        alms,
        L,
        spin=0,
        nside=nside,
        sampling=sampling,
        method=method,
        reality=True,
        precomps=precomps,
        spmd=spmd,
    )
    return f


@partial(
    jax.jit,
    static_argnames=[
        'lmax',
        'mmax',
        'iter',
        'pol',
        'use_weights',
        'datapath',
        'gal_cut',
        'use_pixel_weights',
        'verbose',
        'healpy_ordering',
    ],
)
@requires_s2fft
def map2alm(
    maps,
    lmax=None,
    mmax=None,
    iter=3,
    pol=True,
    use_weights=False,
    datapath=None,
    gal_cut=0,
    use_pixel_weights=False,
    verbose=True,
    healpy_ordering: bool = False,
):
    """Computes the alm of a Healpix map. The input maps must all be
    in ring ordering.

    For recommendations about how to set `lmax`, `iter`, and weights, see the
    `Anafast documentation <https://healpix.sourceforge.io/html/fac_anafast.htm>`_

    Pixel values are weighted before applying the transform:

    * when you don't specify any weights, the uniform weight value 4*pi/n_pix is used
    * with ring weights enabled (use_weights=True), pixels in every ring
      are weighted with a uniform value similar to the one above, ring weights are
      included in healpy
    * with pixel weights (use_pixel_weights=True), every pixel gets an individual weight

    Pixel weights provide the most accurate transform, so you should always use them if
    possible. However they are not included in healpy and will be automatically downloaded
    and cached in ~/.astropy the first time you compute a trasform at a specific nside.

    If datapath is specified, healpy will first check that local folder before downloading
    the weights.
    The easiest way to setup the folder is to clone the healpy-data repository:

    git clone --depth 1 https://github.com/healpy/healpy-data
    cd healpy-data
    bash download_weights_8192.sh

    and set datapath to the root of the repository.

    Parameters
    ----------
    maps : array-like, shape (Npix,) or (n, Npix)
      The input map or a list of n input maps. Must be in ring ordering.
    lmax : int, scalar, optional
      Maximum l of the power spectrum. Default: 3*nside-1
    mmax : int, scalar, optional
      Maximum m of the alm. Default: lmax
    iter : int, scalar, optional
      Number of iteration (default: 3)
    pol : bool, optional
      If True, assumes input maps are TQU. Output will be TEB alm's.
      (input must be 1 or 3 maps)
      If False, apply spin 0 harmonic transform to each map.
      (input can be any number of maps)
      If there is only one input map, it has no effect. Default: True.
    use_weights: bool, scalar, optional
      If True, use the ring weighting. Default: False.
    datapath : None or str, optional
      If given, the directory where to find the pixel weights.
      See in the docstring above details on how to set it up.
    gal_cut : float [degrees]
      pixels at latitude in [-gal_cut;+gal_cut] are not taken into account
    use_pixel_weights: bool, optional
      If True, use pixel by pixel weighting, healpy will automatically download the weights, if needed
    verbose : bool, optional
      Deprecated, has not effect.
    healpy_ordering : bool, optional
      By default, we follow the s2fft ordering for the alms. To use healpy
      ordering, set it to True.

    Returns
    -------
    alms : array or tuple of array
      alm or a tuple of 3 alm (almT, almE, almB) if polarized input.

    Notes
    -----
    The pixels which have the special `UNSEEN` value are replaced by zeros
    before spherical harmonic transform. They are converted back to `UNSEEN`
    value, so that the input maps are not modified. Each map have its own,
    independent mask.
    """
    if mmax is not None:
        raise NotImplementedError('Specifying mmax is not implemented.')
    if iter != 0:
        raise NotImplementedError('Specifying iter > 0 is not implemented')
    if use_weights:
        raise NotImplementedError('Specifying use_weights is not implemented.')
    if datapath is not None:
        raise NotImplementedError('Specifying datapath is not implemented.')
    if gal_cut != 0:
        raise NotImplementedError('Specifying gal_cut is not implemented.')
    if use_pixel_weights:
        raise NotImplementedError('Specifying use_pixel_weights is not implemented.')
    if maps.ndim == 0:
        raise ValueError('The input map must have at least one dimension.')
    if maps.ndim > 2:
        raise ValueError('The input map has too many dimensions.')
    if maps.ndim > 1:
        if pol:
            raise NotImplementedError('TQU maps are not implemented.')
        return jax.vmap(map2alm, in_axes=(0,) + 10 * (None,))(
            maps,
            lmax,
            mmax,
            iter,
            pol,
            use_weights,
            datapath,
            gal_cut,
            use_pixel_weights,
            False,
            healpy_ordering,
        )

    maps = jnp.asarray(maps)
    nside = npix2nside(maps.shape[-1])
    if lmax is None:
        L = 3 * nside
    else:
        L = lmax + 1

    sampling = 'healpix'
    method = 'jax'
    spmd = False

    precomps = generate_precomputes_jax(L, 0, sampling, nside, True)
    flm = spherical.forward(
        maps,
        L,
        spin=0,
        nside=nside,
        sampling=sampling,
        method=method,
        reality=True,
        precomps=precomps,
        spmd=spmd,
    )

    if healpy_ordering:
        return flm_2d_to_hp_fast(flm, L)

    return flm



# This file is part of jax-healpy.
# Copyright (C) 2024 CNRS / SciPol developers
#
# jax-healpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jax-healpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jax-healpy. If not, see <https://www.gnu.org/licenses/>.

from jax import config as _config

from ._query_disc import query_disc
from .clustering._clustering import (
    combine_masks,
    from_cutout_to_fullmap,
    get_clusters,
    get_cutout_from_mask,
    normalize_by_first_occurrence,
    shuffle_labels,
)
from .clustering._kmeans import KMeans, kmeans_sample
from .pixelfunc import (
    UNSEEN,
    ang2pix,
    ang2vec,
    get_interp_val,
    get_interp_weights,
    isnpixok,
    isnsideok,
    maptype,
    nest2ring,
    npix2nside,
    npix2order,
    nside2npix,
    nside2order,
    nside2pixarea,
    nside2resol,
    order2npix,
    order2nside,
    pix2ang,
    pix2vec,
    pix2xyf,
    reorder,
    ring2nest,
    vec2ang,
    vec2pix,
    xyf2pix,
)
from .sphtfunc import alm2map, map2alm

__all__ = [
    'UNSEEN',
    'pix2ang',
    'ang2pix',
    'pix2xyf',
    'xyf2pix',
    'pix2vec',
    'vec2pix',
    'ang2vec',
    'vec2ang',
    'get_interp_weights',
    'get_interp_val',
    # 'get_all_neighbours',
    # 'max_pixrad',
    'nest2ring',
    'ring2nest',
    'reorder',
    # 'ud_grade',
    # 'UNSEEN',
    # 'mask_good',
    # 'mask_bad',
    # 'ma',
    # 'fit_dipole',
    # 'remove_dipole',
    # 'fit_monopole',
    # 'remove_monopole',
    'nside2npix',
    'npix2nside',
    'nside2order',
    'order2nside',
    'order2npix',
    'npix2order',
    'nside2resol',
    'nside2pixarea',
    'isnsideok',
    'isnpixok',
    # 'get_map_size',
    # 'get_min_valid_nside',
    # 'get_nside',
    'maptype',
    # 'ma_to_array',
    'query_disc',
    'alm2map',
    'map2alm',
    # Clustering
    'get_clusters',
    'get_cutout_from_mask',
    'from_cutout_to_fullmap',
    'kmeans_sample',
    'KMeans',
    'combine_masks',
    'normalize_by_first_occurrence',
    'shuffle_labels',
]

_config.update('jax_enable_x64', True)



from pathlib import Path

import numpy as np
import pytest


@pytest.fixture(scope='session')
def numpy_rng() -> np.random.RandomState:
    seed = 0
    return np.random.RandomState(seed)


@pytest.fixture(scope='session')
def data_path() -> Path:
    return Path(__file__).parent / 'data'



# from .._query_disc import boundaries
# from .._pixelfunc import ringinfo, pix2ring, isnsideok
import jax.numpy as jnp
import numpy as np
import pytest

import jax_healpy as hp


def test_nside2npix() -> None:
    assert hp.nside2npix(512) == 3145728
    assert hp.nside2npix(1024) == 12582912


def test_nside2npix_array() -> None:
    assert (hp.nside2npix(np.array([512, 1024])) == np.array([3145728, 12582912])).all()


def test_npix2nside() -> None:
    assert hp.npix2nside(3145728) == 512
    assert hp.npix2nside(12582912) == 1024


@pytest.mark.parametrize('order', range(30))
def test_nside2order_scalar(order: int) -> None:
    assert hp.nside2order(2**order) == order


@pytest.mark.parametrize('order', range(30))
def test_order2nside_scalar(order: int) -> None:
    assert hp.order2nside(order) == 2**order


def test_order2nside_array() -> None:
    o = np.arange(30)
    assert (hp.order2nside(o) == 2**o).all()


def test_nside2resol() -> None:
    np.testing.assert_array_almost_equal(hp.nside2resol(512, arcmin=True), 6.87097282363)
    np.testing.assert_array_almost_equal(hp.nside2resol(1024, arcmin=True), 3.43548641181)


def test_nside2resol_array() -> None:
    np.testing.assert_array_almost_equal(
        hp.nside2resol(np.array([512, 1024]), arcmin=True),
        np.array([6.87097282363, 3.43548641181]),
    )


def test_nside2pixarea() -> None:
    np.testing.assert_array_almost_equal(hp.nside2pixarea(512), 3.9947416351188569e-06)


def test_nside2pixarea_array() -> None:
    np.testing.assert_array_almost_equal(
        hp.nside2pixarea(np.array([512, 512])),
        np.array([3.9947416351188569e-06, 3.9947416351188569e-06]),
    )


@pytest.mark.xfail(reason='max_pixrad not implemented')
def test_max_pixrad() -> None:
    np.testing.assert_array_almost_equal(hp.max_pixrad(512), 2.0870552355e-03)
    np.testing.assert_array_almost_equal(hp.max_pixrad(512, degrees=True), np.rad2deg(2.0870552355e-03))


def test_vec2pix_lonlat(lon0: float, lat0: float) -> None:
    # Need to decrease the precision of the check because deg not radians
    vec = hp.ang2vec(lon0, lat0, lonlat=True)
    lon1, lat1 = hp.vec2ang(vec, lonlat=True)
    np.testing.assert_array_almost_equal(lon1, lon0, decimal=5)
    np.testing.assert_array_almost_equal(lat1, lat0, decimal=5)


def test_get_interp_val_lonlat(theta0: float, phi0: float, lon0: float, lat0: float) -> None:
    m = jnp.arange(12.0)
    val0 = hp.get_interp_val(m, theta0, phi0)
    val1 = hp.get_interp_val(m, lon0, lat0, lonlat=True)
    np.testing.assert_array_almost_equal(val0, val1)


def test_get_interp_weights() -> None:
    p0, w0 = (np.array([0, 1, 4, 5]), np.array([1.0, 0.0, 0.0, 0.0]))

    # phi not specified, theta assumed to be pixel
    p1, w1 = hp.get_interp_weights(1, 0)
    np.testing.assert_array_almost_equal(p0, p1)
    np.testing.assert_array_almost_equal(w0, w1)

    # If phi is not specified, lonlat should do nothing
    p1, w1 = hp.get_interp_weights(1, 0, lonlat=True)
    np.testing.assert_array_almost_equal(p0, p1)
    np.testing.assert_array_almost_equal(w0, w1)

    p0, w0 = (np.array([1, 2, 3, 0]), np.array([0.25, 0.25, 0.25, 0.25]))

    p1, w1 = hp.get_interp_weights(1, 0, 0)
    np.testing.assert_array_almost_equal(p0, p1)
    np.testing.assert_array_almost_equal(w0, w1)

    p1, w1 = hp.get_interp_weights(1, 0, 90, lonlat=True)
    np.testing.assert_array_almost_equal(p0, p1)
    np.testing.assert_array_almost_equal(w0, w1)


@pytest.mark.xfail(reason='get_all_neighbours not implemented')
def test_get_all_neighbours() -> None:
    ipix0 = np.array([8, 4, 0, -1, 1, 6, 9, -1])
    ipix1 = hp.get_all_neighbours(1, np.pi / 2, np.pi / 2)
    ipix2 = hp.get_all_neighbours(1, 90, 0, lonlat=True)
    np.testing.assert_array_almost_equal(ipix0, ipix1)
    np.testing.assert_array_almost_equal(ipix0, ipix2)


@pytest.mark.xfail(reason='fit_dipole not implemented')
def test_fit_dipole() -> None:
    nside = 32
    npix = hp.nside2npix(nside)
    d = [0.3, 0.5, 0.2]
    vec = np.transpose(hp.pix2vec(nside, np.arange(npix)))
    signal = np.dot(vec, d)
    mono, dipole = hp.fit_dipole(signal)
    np.testing.assert_array_almost_equal(mono, 0.0)
    np.testing.assert_array_almost_equal(d[0], dipole[0])
    np.testing.assert_array_almost_equal(d[1], dipole[1])
    np.testing.assert_array_almost_equal(d[2], dipole[2])


@pytest.mark.xfail(reason='boundaries not implemented')
@pytest.mark.parametrize('lg_nside', range(1, 5))
@pytest.mark.parametrize('res', range(1, 50, 7))
@pytest.mark.parametrize('nest', [False, True])
def test_boundaries(lg_nside: int, res: int, nest: bool) -> None:
    """Test whether the boundary shapes look sane"""
    nside = 1 << lg_nside
    for pix in range(hp.nside2npix(nside)):
        num = 4 * res  # Expected number of points
        points = hp.boundaries(nside, pix, res, nest=nest)
        assert points.shape == (3, num)
        dist = np.linalg.norm(points[:, : num - 1] - points[:, 1:])  # distance between points
        assert (dist != 0).all()
        dmin = np.min(dist)
        dmax = np.max(dist)
        assert dmax / dmin <= 2.0


@pytest.mark.xfail(reason='pix2ring not implemented')
@pytest.mark.parametrize('lg_nside', range(1, 5))
@pytest.mark.parametrize('nest', [False, True])
def test_ring(lg_nside: int, nest: bool) -> None:
    nside = 1 << lg_nside
    numPix = hp.nside2npix(nside)
    numRings = 4 * nside - 1  # Expected number of rings
    pix = np.arange(numPix, dtype=np.int64)
    ring = hp.pix2ring(nside, pix, nest=nest)
    assert pix.shape == ring.shape
    assert len(set(ring)) == numRings
    if not nest:
        first = ring[: numPix - 1]
        second = ring[1:]
        assert np.logical_or(first == second, first == second - 1).all()


@pytest.mark.xfail(reason='ud_grade not implemented')
def test_accept_ma_allows_only_keywords() -> None:
    """Test whether the accept_ma wrapper accepts calls using only keywords."""
    ma = np.zeros(12 * 16**2)
    hp.ud_grade(map_in=ma, nside_out=32)


def test_isnsideok() -> None:
    assert hp.isnsideok(nside=1, nest=False)
    assert hp.isnsideok(nside=16, nest=True)

    assert not hp.isnsideok(nside=-16, nest=True)
    assert not hp.isnsideok(nside=-16, nest=False)
    assert not hp.isnsideok(nside=13, nest=True)



import os
import platform
import re
import subprocess
import timeit
from collections.abc import Callable
from dataclasses import asdict, dataclass
from pathlib import Path
from pprint import pprint
from typing import Any, Iterable, Literal

import healpy as hp
import jax
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import typer
import yaml
from jaxtyping import ArrayLike
from matplotlib.ticker import ScalarFormatter

BENCH_PATH = Path(__file__).parent / 'results'
BENCHMARKED_FUNCS = [
    'ang2vec',
    'vec2ang',
    'ang2pix',
    'pix2ang',
    'vec2pix',
    'pix2vec',
    'ring2nest',
    'nest2ring',
    'pix2xyf',
    'xyf2pix',
    'reorder',
    'get_interp_weights',
    'get_interp_val',
]
CHART_PATH_NAME = 'chart-{style}-n{n}.png'

# TODO: use those when typer supports Literals
LibraryType = Literal['healpy', 'jax-healpy']
PrecisionType = Literal['32', '64']

app = typer.Typer()


@dataclass(frozen=True)
class BenchmarkResult:
    library: str
    version: str
    processor: str
    processor_type: str
    precision: int
    n: int
    execution_times_s: dict[str, float]


def bench_it(
    library: str,
    func_name: str,
    nside: int,
    n: int,
    precision: str,
    rng: np.random.Generator,
) -> float:
    if precision == '32':
        dtype = np.dtype('float32')
    elif precision == '64':
        dtype = np.dtype('float64')
    else:
        raise ValueError(f'Invalid precision {precision}')

    args = _get_args(library, func_name, nside, n, dtype, rng)
    func = _get_func(library, func_name, *args)
    with jax.experimental.enable_x64(precision == '64'):
        return time_it(func)


def _get_args(
    library: str,
    func_name: str,
    nside: int,
    n: int,
    dtype: np.dtype,
    rng: np.random.Generator,
) -> tuple[ArrayLike]:
    if func_name in {'ang2vec', 'ang2pix', 'get_interp_weights'}:
        theta = rng.uniform(0, np.pi, size=n).astype(dtype)
        phi = rng.uniform(0, 2 * np.pi, size=n).astype(dtype)
        if func_name == 'ang2vec':
            args = (theta, phi)
        else:
            args = (nside, theta, phi)

    elif func_name == 'get_interp_val':
        # Generate a random map and interpolation points
        npix = 12 * nside**2
        m = rng.uniform(0, 10, size=npix).astype(dtype)
        theta = rng.uniform(0, np.pi, size=n).astype(dtype)
        phi = rng.uniform(0, 2 * np.pi, size=n).astype(dtype)
        args = (m, theta, phi)

    elif func_name in {'vec2ang', 'vec2pix'}:
        vec = rng.uniform(-np.pi, np.pi, size=(3, n))
        vec /= np.sqrt(np.sum(vec**2, axis=0))
        vec = vec.astype(dtype)
        if func_name == 'vec2ang':
            args = (vec.T.copy(),)
        else:
            args = (nside, vec[0], vec[1], vec[2])

    elif func_name in {'pix2ang', 'pix2vec', 'pix2xyf', 'ring2nest', 'nest2ring'}:
        pixels = rng.uniform(0, hp.nside2npix(nside), size=n).astype(int)
        args = (nside, pixels)

    elif func_name == 'xyf2pix':
        x = rng.uniform(0, nside, size=n).astype(int)
        y = rng.uniform(0, nside, size=n).astype(int)
        f = rng.uniform(0, 12, size=n).astype(int)
        args = (nside, x, y, f)

    elif func_name == 'reorder':
        npix = hp.nside2npix(nside)
        map_in = rng.uniform(size=npix)
        args = (map_in,)

    else:
        raise NotImplementedError

    if library == 'jax-healpy':
        args = tuple(jax.device_put(_) if isinstance(_, np.ndarray) else _ for _ in args)

    return args


def _get_func(library: str, func_name: str, *args: Any):
    if library == 'jax-healpy':
        import jax_healpy as module
    else:
        import healpy as module

    func = getattr(module, func_name)
    if library == 'healpy':
        if func_name == 'reorder':
            func_ = lambda: func(*args, r2n=True)  # noqa: E731
        else:
            func_ = lambda: func(*args)  # noqa: E731
    else:
        if func_name in {'pix2ang', 'vec2ang'}:

            def func_() -> None:
                theta, phi = func(*args)
                theta.block_until_ready()
                phi.block_until_ready()

        elif func_name == 'pix2xyf':

            def func_() -> None:
                x, y, f = func(*args)
                x.block_until_ready()
                y.block_until_ready()
                f.block_until_ready()

        elif func_name == 'reorder':

            def func_() -> None:
                func(*args, r2n=True).block_until_ready()

        elif func_name == 'get_interp_weights':

            def func_() -> None:
                p, w = func(*args)
                p.block_until_ready()
                w.block_until_ready()

        elif func_name == 'get_interp_val':

            def func_() -> None:
                result = func(*args)
                result.block_until_ready()

        else:

            def func_() -> None:
                func(*args).block_until_ready()

        func_()  # discard first call, which includes compilation

    return func_


def time_it(func: Callable[[], None]) -> float:
    timer = timeit.Timer(func)
    number, _ = timer.autorange()
    execution_time = min(timer.repeat(number=number)) / number
    return execution_time


@app.command()
def run(
    library: str,
    nside: int = 512,
    n: int = 10_000_000,
    precision: str = '64',
) -> None:
    if library == 'jax-healpy':
        version = f'jax({jax.__version__})'
        device = jax.devices()[0]
        processor = device.device_kind
        if processor == 'cpu':
            processor = _get_cpu_name()
            processor_type = 'cpu'
        else:
            processor_type = 'gpu'
    elif library == 'healpy':
        version = hp.__version__
        processor = _get_cpu_name()
        processor_type = 'cpu'
    else:
        raise ValueError(f'Invalid library {library}')

    rng = np.random.default_rng(0)
    execution_times = {}

    print(f'Running {library}...')
    for func_name in BENCHMARKED_FUNCS:
        execution_time = bench_it(library, func_name, nside, n, precision, rng)
        execution_times[func_name] = execution_time

    result = BenchmarkResult(
        library=library,
        version=version,
        processor=processor,
        processor_type=processor_type,
        precision=int(precision),
        n=n,
        execution_times_s=execution_times,
    )
    pprint(result)

    chars = '[] '
    formatted_processor = processor.lower().rstrip(chars).replace('(c)', '').replace('(tm)', '')
    filename = f'{library}-{formatted_processor}-precision{precision}-n{n}.yaml'
    for char in chars:
        filename = filename.replace(char, '-')
    BENCH_PATH.mkdir(parents=True, exist_ok=True)
    with open(BENCH_PATH / filename, 'w') as f:
        yaml.dump(asdict(result), f)


def _get_cpu_name():
    if platform.system() == 'Darwin':
        os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
        command = 'sysctl -n machdep.cpu.brand_string'
        return subprocess.check_output(command).strip()
    elif platform.system() == 'Linux':
        command = 'cat /proc/cpuinfo'
        all_info = subprocess.check_output(command, shell=True).decode().strip()
        for line in all_info.split('\n'):
            if 'model name' in line:
                return re.sub('.*model name.*:', '', line, 1).strip()
    return platform.processor()


@app.command()
def collect():
    files = BENCH_PATH.glob('*-n*.yaml')
    ns = {}
    for file in files:
        result_as_dict = yaml.safe_load(file.read_text())
        result = BenchmarkResult(**result_as_dict)
        ns.setdefault(result.n, []).append(result)

    for n, results in ns.items():
        results = sorted(
            results,
            key=lambda _: sum(_.execution_times_s.values()) / len(BENCHMARKED_FUNCS),
        )
        export_chart(n, results, 'default')
        export_chart(n, results, 'dark_background')


def export_chart(n: int, results: Iterable[BenchmarkResult], style: str):
    chart_path = BENCH_PATH / CHART_PATH_NAME.format(style=style.replace('_', ''), n=n)
    plt.style.use(style)
    # I've used pandas because I was not able to do what I wanted with matplotlib alone
    df = pd.DataFrame(
        [
            [f'{res.library}[{res.processor_type}{res.precision}]']
            + [res.execution_times_s[_] for _ in BENCHMARKED_FUNCS]
            for res in results
        ],
        columns=['library'] + BENCHMARKED_FUNCS,
    )
    for func in BENCHMARKED_FUNCS:
        df[func] /= df[func][0]
    pprint(df)

    ax = df.plot.bar(x='library', title='Benchmark (lower is better)', rot=60)
    ax.legend(framealpha=0, loc='upper left')
    plt.yscale('log')
    ax.yaxis.set_major_formatter(ScalarFormatter())
    plt.xlabel('')
    plt.tight_layout()
    plt.savefig(chart_path, transparent=True)


if __name__ == '__main__':
    app()



import timeit
from pathlib import Path
from typing import Callable

import healpy as hp
import jax
import numpy as np

import jax_healpy as jhp

data_path = Path(jhp.__file__).parent.parent / 'tests/data'


def cla(data_path: Path) -> np.ndarray:
    return hp.read_cl(data_path / 'cl_wmap_band_iqumap_r9_7yr_W_v4_udgraded32_II_lmax64_rmmono_3iter.fits')


nside = 256
lmax = 2 * nside
fwhm_deg = 7.0


orig = hp.synfast(
    cla(data_path),
    nside,
    lmax=lmax,
    pixwin=False,
    fwhm=np.radians(fwhm_deg),
    new=False,
)
orig_d = jax.device_put(orig)


def map2alm_hp():
    hp.map2alm(orig, iter=0, lmax=2 * nside - 1, use_weights=False)


def map2alm_jhp():
    jhp.map2alm(orig_d).block_until_ready()


def time_it(func: Callable[[], None]) -> float:
    timer = timeit.Timer(func)
    number, _ = timer.autorange()
    execution_time = min(timer.repeat(number=number)) / number
    return execution_time


print(time_it(map2alm_hp))
# 0.0038599122400046326

print(time_it(map2alm_jhp))
# 0.437624677999338



# This file is part of jax-healpy.
# Copyright (C) 2024 CNRS / SciPol developers
#
# jax-healpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jax-healpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jax-healpy. If not, see <https://www.gnu.org/licenses/>.

"""
Query disc implementation for HEALPix spherical disc queries.
"""

from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
from jax import jit, lax
from jaxtyping import Array, ArrayLike

from .pixelfunc import pix2vec

__all__ = ['query_disc']


def _compute_resolution(nside: int) -> Array:
    """JAX-compatible computation of pixel resolution in radians.

    This computes the approximate pixel size as the square root of the pixel area.
    This is a gross approximation given the different pixel shapes in HEALPix.

    Parameters
    ----------
    nside : int
        HEALPix nside parameter

    Returns
    -------
    resol : Array
        Approximate pixel size in radians
    """
    npix = 12 * nside * nside  # Total number of pixels
    pixarea = 4 * jnp.pi / npix  # Pixel area = total sphere area / npix
    return jnp.sqrt(pixarea)  # Resolution = sqrt(area)


@partial(jit, static_argnames=['nside', 'inclusive', 'fact', 'nest', 'max_length'])
def query_disc(
    nside: int,
    vec: ArrayLike,
    radius: float,
    inclusive: bool = False,
    fact: int = 4,
    nest: bool = False,
    max_length: Optional[int] = None,
) -> Array:
    """Find pixels within a disc on the sphere.

    This function supports both single and batched queries. It is fully JIT-compatible
    and differentiable with respect to vec and radius parameters.

    Parameters
    ----------
    nside : int
        The resolution parameter of the HEALPix map
    vec : array-like
        Either a single three-component unit vector (3,) defining the center of the disc,
        or a batch of vectors (3, B) defining B disc centers
    radius : float
        The radius of the disc in radians
    inclusive : bool, optional
        If False (default), return pixels whose centers lie within the disc.
        Results are guaranteed to match healpy exactly for inclusive=False.
        If True, return all pixels that overlap with the disc.
        Note: inclusive=True may produce slightly different results compared to healpy
        due to algorithm differences in determining pixel overlap.
    fact : int, optional
        For inclusive queries, the pixelization factor (default: 4)
    nest : bool, optional
        If True, assume NESTED pixel ordering, otherwise RING ordering (default: False)
    max_length : int, optional
        Maximum number of pixels to return per disc. If None, defaults to npix.
        For batched inputs, this limits memory usage by returning (B, max_length)
        instead of (B, npix).

    Returns
    -------
    ipix : array
        For single vector input (3,): returns (max_length,) or (npix,) if max_length is None
        For batch vector input (3, B): returns (B, max_length)
        Pixels outside the disc are marked as npix (sentinel value).
        This allows direct indexing like: map.at[disc].set(value)

    Raises
    ------
    NotImplementedError
        If nest=True (nested ordering not yet supported)

    Notes
    -----
    This function currently only supports RING ordering. The function is
    JIT-compatible and differentiable when compiled with static_argnums=(0,)
    for the nside parameter. The returned array has fixed size for JIT
    compatibility - pixels outside the disc have value npix.

    When indexing a JAX array, pixels outside the disc should be ignored.
    If indexing a numpy array, this will raise an out-of-bounds error.

    Examples
    --------
    Single disc query:

    >>> import jax_healpy as hp
    >>> import jax.numpy as jnp
    >>> import jax
    >>> nside = 16
    >>> vec = jnp.array([1.0, 0.0, 0.0])  # Point on equator
    >>> radius = 0.1  # ~5.7 degrees
    >>> disc = hp.query_disc(nside, vec, radius)
    >>> # Use directly for indexing
    >>> map = jnp.zeros(hp.nside2npix(nside))
    >>> map = map.at[disc].set(1.0)

    Batch disc query:

    >>> vecs = jnp.array([[1.0, 0.0], [0.0, 1.0], [0.0, 0.0]])  # (3, 2) - two centers
    >>> discs = hp.query_disc(nside, vecs, radius, max_length=1000)  # (2, 1000)
    >>> # Each row contains pixels for one disc
    >>> map1 = map.at[discs[0]].set(1.0)  # First disc
    >>> map2 = map.at[discs[1]].set(2.0)  # Second disc

    For JIT compilation:

    >>> jit_query_disc = jax.jit(
    ...     lambda n, v, r: hp.query_disc(n, v, r)
    ... )
    >>> disc_jit = jit_query_disc(nside, vec, radius)
    """
    # Raise error for nested ordering
    if nest:
        raise NotImplementedError('Nested ordering not yet supported')

    return _query_disc_ring(nside, vec, radius, inclusive, fact, max_length)


def _query_disc_ring(
    nside: int, vec: ArrayLike, radius: float, inclusive: bool, fact: int, max_length: Optional[int]
) -> Array:
    """JIT-compatible unified implementation of disc query supporting both single and batch inputs.

    Algorithm Overview:
    1. Standardize input to (3, batch_dims) format and set defaults
    2. Normalize input vectors and clip radius to valid range
    3. Calculate the cosine threshold for the dot product test
    4. Generate all pixel vectors and compute broadcast dot products
    5. Create mask for pixels within the disc(s)
    6. Select top max_length pixels per disc with sentinel padding
    7. Apply JAX-compatible warning system for truncation
    8. Squeeze output for single vector compatibility
    """
    # Step 1: Input standardization to (3, batch_dims) format
    vec = jnp.asarray(vec, dtype=jnp.float64)
    original_is_single = vec.ndim == 1
    if original_is_single:
        vec = vec[:, None]  # (3,) → (3, 1)

    batch_dims = vec.shape[1]
    npix = 12 * nside * nside
    radius = jnp.asarray(radius, dtype=jnp.float64)

    # Default max_length to npix if not provided
    if max_length is None:
        max_length = npix

    # Step 2: Normalize center vectors (handle zero vector case)
    vec_norms = jnp.linalg.norm(vec, axis=0)  # (batch_dims,)
    # Create default direction - broadcasts to (3, batch_dims)
    default_dir = jnp.array([1.0, 0.0, 0.0])[:, None]

    # Normalize each vector individually
    safe_vecs = jnp.where(vec_norms[None, :] > 1e-10, vec / vec_norms[None, :], default_dir)

    # Clip radius to valid range [0, π]
    radius = jnp.clip(radius, 0.0, jnp.pi)

    # Step 3: Calculate cosine threshold for dot product comparison
    cos_radius = jnp.cos(radius)

    # For inclusive mode, expand the radius by pixel resolution divided by fact
    if inclusive:
        expanded_radius = radius + _compute_resolution(nside) / fact
        cos_expanded_radius = jnp.cos(jnp.clip(expanded_radius, 0, jnp.pi))
    else:
        cos_expanded_radius = cos_radius

    # Step 4: Generate all pixel vectors and compute broadcast dot products
    all_pixels = jnp.arange(npix, dtype=jnp.int32)
    pixel_vecs = pix2vec(nside, all_pixels, nest=False)  # (npix, 3)

    # Broadcast dot products: (npix, 3) @ (3, batch_dims) → (npix, batch_dims)
    dot_products = jnp.dot(pixel_vecs, safe_vecs)

    # Step 5: Create mask for pixels within the disc(s)
    # Use small tolerance to handle floating point precision issues
    tolerance = 1e-6
    mask = dot_products >= (cos_expanded_radius - tolerance)  # (npix, batch_dims)

    # Step 6: Select top max_length pixels per disc
    # Create sort keys: valid pixels keep dot product, invalid get -inf
    sort_keys = jnp.where(mask, dot_products, -jnp.inf)  # (npix, batch_dims)

    # Sort indices by dot product (best pixels last)
    sorted_indices = jnp.argsort(sort_keys, axis=0)  # (npix, batch_dims)

    # Select top max_length pixels per batch
    top_indices = sorted_indices[-max_length:]  # (max_length, batch_dims)
    result = top_indices.T  # (batch_dims, max_length)

    # Replace invalid entries (where sort key was -inf) with npix
    selected_scores = sort_keys[top_indices, jnp.arange(batch_dims)]  # (max_length, batch_dims)
    invalid_mask = selected_scores.T == -jnp.inf  # (batch_dims, max_length)
    result = jnp.where(invalid_mask, npix, result)

    # Step 7: JAX-compatible warning system for truncation
    if max_length < npix:  # Only check for truncation if limiting
        # Count valid pixels per batch
        valid_counts = jnp.sum(mask.astype(jnp.int32), axis=0)  # (batch_dims,)
        exceeded_count = jnp.sum(valid_counts > max_length)  # scalar

        # Use lax.cond for JAX-compatible conditional warning
        lax.cond(
            exceeded_count > 0,
            lambda: jax.debug.print('Warning: {} valid pixels exceeded max_length={}', valid_counts.max(), max_length),
            lambda: None,
        )

    # Step 8: Squeeze output for single vector compatibility
    if original_is_single:
        result = jnp.squeeze(result, axis=0)  # (1, max_length) → (max_length,)

    return result



import healpy as hp
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import jax_healpy as jhp


def test_basic_functionality():
    """Test basic functionality of query_disc."""
    nside = 16
    vec = [1.0, 0.0, 0.0]  # Point on equator
    radius = 0.1  # ~5.7 degrees

    # Test basic call
    ipix = jhp.query_disc(nside, vec, radius)
    assert len(ipix) == jhp.nside2npix(nside), 'Should return fixed-size array'

    # Count valid pixels (those < npix)
    npix = jhp.nside2npix(nside)
    valid_pixels = ipix[ipix < npix]
    assert len(valid_pixels) > 0, 'Should find some pixels'

    # Test with different parameters
    ipix_inclusive = jhp.query_disc(nside, vec, radius, inclusive=True)
    valid_pixels_inclusive = ipix_inclusive[ipix_inclusive < npix]
    assert len(valid_pixels_inclusive) >= len(valid_pixels), 'Inclusive should return more or equal pixels'


def test_map_indexing_example():
    """Test the map indexing example provided by user."""
    nside = 16

    # JAX-healpy version
    jhp_map = jnp.zeros((jhp.nside2npix(nside),), dtype=jnp.float32)
    central_pix = jhp.ang2pix(nside, np.pi / 2, 0.0, lonlat=False)
    central_vec = jhp.pix2vec(nside, central_pix)
    disc = jhp.query_disc(nside, central_vec, np.pi / 4, inclusive=True)
    jhp_map = jhp_map.at[disc].set(1.0)

    # healpy version
    hp_map = np.zeros((hp.nside2npix(nside),), dtype=np.float32)
    central_pix_hp = hp.ang2pix(nside, np.pi / 2, 0.0, lonlat=False)
    central_vec_hp = hp.pix2vec(nside, central_pix_hp)
    disc_hp = hp.query_disc(nside, central_vec_hp, np.pi / 4, inclusive=True)
    hp_map[disc_hp] = 1.0

    # Compare the maps - they should be reasonably similar
    # Allow for some differences in inclusive mode due to algorithm differences
    jhp_marked = np.where(jhp_map > 0.5)[0]
    hp_marked = np.where(hp_map > 0.5)[0]

    # Check overlap ratio rather than exact match for inclusive mode
    overlap = len(set(jhp_marked) & set(hp_marked))
    total_unique = len(set(jhp_marked) | set(hp_marked))
    overlap_ratio = overlap / total_unique if total_unique > 0 else 1.0
    assert overlap_ratio > 0.7, f'Insufficient map overlap: {overlap_ratio:.2f}'


@pytest.mark.parametrize(
    'nside,vec,radius,inclusive,description',
    [
        (16, [1.0, 0.0, 0.0], 0.1, False, 'Small disc on equator'),
        (16, [0.0, 0.0, 1.0], 0.2, False, 'Small disc at north pole'),
        (16, [0.0, 0.0, -1.0], 0.15, False, 'Small disc at south pole'),
        (32, [1.0, 0.0, 0.0], 0.05, False, 'Very small disc'),
        (8, [0.5, 0.5, 0.707], 0.3, False, 'Medium disc at arbitrary position'),
        (16, [1.0, 0.0, 0.0], 0.1, True, 'Small disc inclusive mode'),
        (16, [0.0, 0.0, 1.0], 0.2, True, 'Pole disc inclusive mode'),
    ],
)
def test_against_healpy_comprehensive(nside, vec, radius, inclusive, description):
    """Comprehensive test against healpy implementation."""
    # JAX-healpy result
    jax_result = jhp.query_disc(nside, vec, radius, inclusive=inclusive)

    # healpy result
    healpy_result = hp.query_disc(nside, vec, radius, inclusive=inclusive)

    # Extract valid pixels from jax result
    npix = jhp.nside2npix(nside)
    jax_valid = jax_result[jax_result < npix]

    # Sort both results for comparison
    jax_sorted = np.sort(jax_valid)
    healpy_sorted = np.sort(healpy_result)

    # Compare results - allow for some differences in inclusive mode
    if inclusive:
        # For inclusive mode, check that results are reasonably close
        overlap = len(set(jax_sorted) & set(healpy_sorted))
        total_unique = len(set(jax_sorted) | set(healpy_sorted))
        overlap_ratio = overlap / total_unique if total_unique > 0 else 1.0
        assert overlap_ratio > 0.7, f'Insufficient overlap for {description}: {overlap_ratio:.2f}'
    else:
        # For non-inclusive mode, expect exact match
        np.testing.assert_array_equal(jax_sorted, healpy_sorted, err_msg=f'Failed for: {description}')


def test_edge_cases():
    """Test edge cases and special conditions."""
    nside = 16
    vec = [1.0, 0.0, 0.0]
    npix = jhp.nside2npix(nside)

    # Test with zero radius
    ipix = jhp.query_disc(nside, vec, 0.0)
    valid_pixels = ipix[ipix < npix]
    # Zero radius correctly returns no pixels (both JAX-healpy and healpy)
    assert len(valid_pixels) == 0, 'Zero radius should return no pixels'

    # Test with very large radius (full sphere)
    ipix = jhp.query_disc(nside, vec, np.pi)
    valid_pixels = ipix[ipix < npix]
    assert len(valid_pixels) == npix, 'Full sphere should return all pixels'

    # Test with non-unit vector (should be normalized)
    vec_unnorm = [2.0, 0.0, 0.0]
    ipix1 = jhp.query_disc(nside, vec, 0.1)
    ipix2 = jhp.query_disc(nside, vec_unnorm, 0.1)
    np.testing.assert_array_equal(ipix1, ipix2, 'Non-unit vector should be normalized')


def test_input_validation():
    """Test input validation and error conditions."""
    nside = 16
    vec = [1.0, 0.0, 0.0]

    # Test zero vector (should be handled gracefully)
    ipix = jhp.query_disc(nside, [0.0, 0.0, 0.0], 0.1)
    assert len(ipix) == jhp.nside2npix(nside), 'Zero vector should be handled'

    # Test negative radius (should be clipped)
    ipix = jhp.query_disc(nside, vec, -0.1)
    assert len(ipix) == jhp.nside2npix(nside), 'Negative radius should be clipped'


def test_nested_ordering_error():
    """Test that nested ordering raises NotImplementedError."""
    nside = 16
    vec = [1.0, 0.0, 0.0]
    radius = 0.1

    with pytest.raises(NotImplementedError, match='Nested ordering not yet supported'):
        jhp.query_disc(nside, vec, radius, nest=True)


def test_jit_compatibility():
    """Test that the function is JIT-compatible."""
    import jax

    nside = 16
    vec = jnp.array([1.0, 0.0, 0.0])
    radius = 0.1

    # JIT compile the function with nside as static argument
    jit_query_disc = jax.jit(lambda n, v, r: jhp.query_disc(n, v, r), static_argnums=(0,))

    # Test that it works
    result = jit_query_disc(nside, vec, radius)
    assert len(result) == jhp.nside2npix(nside), 'JIT-compiled function should work'

    # Test that multiple calls work (compilation caching)
    result2 = jit_query_disc(nside, vec, radius)
    np.testing.assert_array_equal(result, result2, 'JIT-compiled function should be deterministic')


def test_differentiability():
    """Test that the function is differentiable w.r.t. vec and radius."""

    nside = 8  # Small nside for faster computation
    vec = jnp.array([1.0, 0.0, 0.0])
    radius = 0.2

    # Test gradient w.r.t. vec
    def loss_vec(v):
        # Use JIT with static nside for performance
        jit_query = jax.jit(lambda n, v, r: jhp.query_disc(n, v, r), static_argnums=(0,))
        disc = jit_query(nside, v, radius)
        npix = jhp.nside2npix(nside)
        valid_count = jnp.sum(disc < npix)
        return valid_count.astype(jnp.float32)

    grad_vec = jax.grad(loss_vec)(vec)
    assert grad_vec.shape == (3,), 'Gradient w.r.t. vec should have shape (3,)'

    # Test gradient w.r.t. radius
    def loss_radius(r):
        # Use JIT with static nside for performance
        jit_query = jax.jit(lambda n, v, r: jhp.query_disc(n, v, r), static_argnums=(0,))
        disc = jit_query(nside, vec, r)
        npix = jhp.nside2npix(nside)
        valid_count = jnp.sum(disc < npix)
        return valid_count.astype(jnp.float32)

    grad_radius = jax.grad(loss_radius)(radius)
    assert isinstance(grad_radius, jnp.ndarray), 'Gradient w.r.t. radius should be computed'


@pytest.mark.parametrize('radius', [0.01, 0.1, 0.5, 1.0])
def test_fixed_size_output(radius):
    """Test that output has fixed size with npix sentinel values."""
    nside = 16
    vec = [1.0, 0.0, 0.0]
    npix = jhp.nside2npix(nside)

    result = jhp.query_disc(nside, vec, radius)
    assert len(result) == npix, f'Output should have fixed size {npix}'

    # Check that sentinel values are npix
    invalid_mask = result == npix
    valid_mask = result < npix
    assert np.all(result[invalid_mask] == npix), 'Invalid pixels should have value npix'
    assert np.all(result[valid_mask] < npix), 'Valid pixels should have value < npix'


def test_batch_functionality():
    """Test batched input functionality with multiple disc centers."""
    nside = 16
    npix = jhp.nside2npix(nside)
    radius = 0.1
    max_length = 100

    # Test batch of 3 vectors
    vecs = jnp.array(
        [
            [1.0, 0.0, 0.0],  # equator
            [0.0, 1.0, 0.0],  # equator, 90 degrees rotated
            [0.0, 0.0, 1.0],  # north pole
        ]
    ).T  # Shape (3, 3)

    # Test batch query
    batch_result = jhp.query_disc(nside, vecs, radius, max_length=max_length)
    assert batch_result.shape == (3, max_length), f'Expected (3, {max_length}), got {batch_result.shape}'

    # Test individual queries for comparison
    for i in range(3):
        single_vec = vecs[:, i]
        single_result = jhp.query_disc(nside, single_vec, radius, max_length=max_length)

        # Get valid pixels from both
        batch_valid = batch_result[i][batch_result[i] < npix]
        single_valid = single_result[single_result < npix]

        # Should have same valid pixels (sorted)
        np.testing.assert_array_equal(
            np.sort(batch_valid),
            np.sort(single_valid),
            err_msg=f"Batch result for vector {i} doesn't match single result",
        )

    # Test that invalid pixels are marked as npix
    invalid_mask = batch_result == npix
    valid_mask = batch_result < npix
    assert np.all(batch_result[invalid_mask] == npix), 'Invalid pixels should be npix'
    assert np.all(batch_result[valid_mask] < npix), 'Valid pixels should be < npix'

    # Test with single vector (should squeeze back)
    single_vec = vecs[:, 0]
    single_result = jhp.query_disc(nside, single_vec, radius, max_length=max_length)
    assert single_result.shape == (max_length,), f'Single vector should return ({max_length},)'

    # Compare with first batch element
    batch_first = batch_result[0]
    np.testing.assert_array_equal(single_result, batch_first, 'Single vector result should match first batch element')


def test_batch_edge_cases():
    """Test edge cases for batch functionality."""
    nside = 8  # Small for faster testing
    npix = jhp.nside2npix(nside)

    # Test with very small max_length (truncation)
    vecs = jnp.array([[1.0, 0.0], [0.0, 1.0], [0.0, 0.0]])  # (3, 2)
    max_length = 5

    result = jhp.query_disc(nside, vecs, 0.5, max_length=max_length)  # Large radius
    assert result.shape == (2, max_length), f'Expected (2, {max_length})'

    # Should have valid pixels followed by npix padding
    for i in range(2):
        row = result[i]
        valid_pixels = row[row < npix]
        assert len(valid_pixels) <= max_length, 'Should not exceed max_length valid pixels'

    # Test with max_length larger than typical disc size
    large_max_length = npix // 2
    result2 = jhp.query_disc(nside, vecs, 0.1, max_length=large_max_length)  # Small radius
    assert result2.shape == (2, large_max_length), f'Expected (2, {large_max_length})'

    # Most entries should be npix (padding)
    for i in range(2):
        row = result2[i]
        valid_pixels = row[row < npix]
        padding_pixels = row[row == npix]
        assert len(valid_pixels) + len(padding_pixels) == large_max_length
        assert len(valid_pixels) < large_max_length  # Should have some padding



"""Tests for get_interp_weights function."""

import healpy as hp
import jax
import jax.numpy as jnp
import numpy as np
import pytest
from numpy.testing import assert_allclose

import jax_healpy as jhp


@pytest.mark.parametrize('region_name', ['Low Cap', 'Equator', 'High Cap', 'All'])
@pytest.mark.parametrize('nside', [4, 8, 16, 32, 64, 128, 256])
def test_get_interp_weights_regional_precision(region_name, nside):
    """Test precision in different HEALPix regions based on playground validation."""
    npix = jhp.nside2npix(nside)

    # Define HEALPix regions exactly like playground
    low_cap_start, low_cap_end = 0, 2 * nside * (nside - 1)
    equator_start, equator_end = low_cap_end, npix - low_cap_end
    high_cap_start, high_cap_end = equator_end, npix

    region_map = {
        'Low Cap': (low_cap_start, low_cap_end),
        'Equator': (equator_start, equator_end),
        'High Cap': (high_cap_start, high_cap_end),
        'All': (0, npix),
    }

    start, end = region_map[region_name]

    # Test all pixels in region (like playground)
    ipix = jnp.arange(start, end, dtype=jnp.int32)

    # Get base coordinates and add small perturbations (same as playground)
    theta, phi = jhp.pix2ang(nside, ipix)
    theta_perturb = theta + jax.random.uniform(jax.random.key(0), theta.shape, minval=-0.01, maxval=0.01)
    phi_perturb = phi + jax.random.uniform(jax.random.key(1), phi.shape, minval=-0.01, maxval=0.01)

    # Get interpolation weights (same as playground)
    pixels, weights = jhp.get_interp_weights(nside, theta_perturb, phi_perturb)
    hp_pixels, hp_weights = hp.get_interp_weights(nside, theta_perturb, phi_perturb)

    # Sort for comparison (same as playground)
    sorted_indices = jnp.argsort(pixels, axis=0)
    sorted_pixels = jnp.take_along_axis(pixels, sorted_indices, axis=0)
    sorted_weights = jnp.take_along_axis(weights, sorted_indices, axis=0)

    sorted_hp_indices = np.argsort(hp_pixels, axis=0)
    sorted_hp_pixels = np.take_along_axis(hp_pixels, sorted_hp_indices, axis=0)
    sorted_hp_weights = np.take_along_axis(hp_weights, sorted_hp_indices, axis=0)

    # Calculate weight error for validation
    weight_error = jnp.mean((sorted_weights - sorted_hp_weights) ** 2)

    # Focus on weight precision - this is what the user cares about
    # The playground shows that weights sum to 1 and have small errors

    # First verify weights sum to 1 (fundamental requirement)
    weight_sums = jnp.sum(weights, axis=0)
    assert_allclose(weight_sums, 1.0, rtol=1e-12, atol=1e-12, err_msg=f'{region_name} weights should sum to 1.0')
    # Check pixel and weight errors to be very small

    # Ultra-strict precision testing - as strict as numerically achievable
    # Based on empirical testing of actual precision limits per nside
    if nside <= 64:
        # Demand near machine precision for low nside values
        assert weight_error < 1e-25, (
            f'{region_name} region weight MSE {weight_error} should achieve ultra precision for nside {nside}'
        )
    else:
        # nside=256: precision is limited by phi interpolation algorithm differences with healpy
        # Set thresholds based on empirical observations with safety margin
        precision_threshold = 1e-25 if region_name == 'Equator' else 1e-5
        assert weight_error < precision_threshold, f"""
            {region_name} region weight MSE {weight_error} exceeds maximum achievable precision for nside
            {nside} (threshold {precision_threshold:.2e})
            """

    # Verify all weights are non-negative (physical requirement)
    assert jnp.all(weights >= 0), f'{region_name} all weights should be non-negative'

    # Pixel accuracy validation - as strict as possible while accounting for known algorithmic limitations
    max_pixel_diff = jnp.max(jnp.abs(sorted_pixels - sorted_hp_pixels))

    # Set pixel difference thresholds based on empirical testing and known precision limits
    if nside <= 128:
        # For nside <= 128, demand perfect pixel matching in most cases
        max_allowed_diff = 0
        assert max_pixel_diff == max_allowed_diff, (
            f'{region_name} region nside {nside}: demand pixel-exact matching, found max diff {max_pixel_diff}'
        )
    else:
        # For nside >= 256, allow minimal pixel differences due to phi interpolation algorithm limitations
        # These differences occur in challenging coordinate regions (poles, ring transitions)
        # But still maintain strict bounds - much better than original test
        max_allowed_diff = 4 if region_name == 'All' else 0  # Matches the observed maximum difference for nside=256
        assert max_pixel_diff <= max_allowed_diff, f"""
            {region_name} region nside {nside}: max pixel diff {max_pixel_diff}
            exceeds strict algorithmic limit {max_allowed_diff}.
            This indicates a precision regression beyond known limitations.
            """


@pytest.mark.slow
@pytest.mark.parametrize('region_name', ['Low Cap', 'Equator', 'High Cap', 'All'])
@pytest.mark.parametrize('nside', [512, 1024, 2048, 4096, 8192])
def test_get_interp_weights_high_nside_sampling(region_name, nside):
    """Test precision for high nside values using random sampling to avoid memory issues."""
    npix = jhp.nside2npix(nside)

    # Define HEALPix regions exactly like the full precision test
    low_cap_start, low_cap_end = 0, 2 * nside * (nside - 1)
    equator_start, equator_end = low_cap_end, npix - low_cap_end
    high_cap_start, high_cap_end = equator_end, npix

    region_map = {
        'Low Cap': (low_cap_start, low_cap_end),
        'Equator': (equator_start, equator_end),
        'High Cap': (high_cap_start, high_cap_end),
        'All': (0, npix),
    }

    start, end = region_map[region_name]
    region_size = end - start

    # Use random sampling to test substantial number of pixels from the region
    # Increase sample size for more comprehensive testing while remaining memory-efficient
    n_samples = min(5000, region_size)  # Test up to 5000 pixels per region for thorough coverage

    # Generate random pixel indices within the region
    key = jax.random.key(11)  # Fixed seed for reproducibility
    if region_size > n_samples:
        # Sample random indices within the region
        random_offsets = jax.random.choice(key, region_size, (n_samples,), replace=False)
        ipix = start + random_offsets
    else:
        # If region is small, test all pixels
        ipix = jnp.arange(start, end, dtype=jnp.int32)

    # Get base coordinates and add small perturbations
    theta, phi = jhp.pix2ang(nside, ipix)
    key1, key2 = jax.random.split(key, 2)
    theta_perturb = theta + jax.random.uniform(key1, theta.shape, minval=-0.01, maxval=0.01)
    phi_perturb = phi + jax.random.uniform(key2, phi.shape, minval=-0.01, maxval=0.01)

    # Get interpolation weights
    pixels, weights = jhp.get_interp_weights(nside, theta_perturb, phi_perturb)
    hp_pixels, hp_weights = hp.get_interp_weights(nside, theta_perturb, phi_perturb)

    # Sort for comparison
    sorted_indices = jnp.argsort(pixels, axis=0)
    sorted_pixels = jnp.take_along_axis(pixels, sorted_indices, axis=0)
    sorted_weights = jnp.take_along_axis(weights, sorted_indices, axis=0)

    sorted_hp_indices = np.argsort(hp_pixels, axis=0)
    sorted_hp_pixels = np.take_along_axis(hp_pixels, sorted_hp_indices, axis=0)
    sorted_hp_weights = np.take_along_axis(hp_weights, sorted_hp_indices, axis=0)

    # Calculate weight error for validation
    weight_error = jnp.mean((sorted_weights - sorted_hp_weights) ** 2)

    # First verify weights sum to 1 (fundamental requirement)
    weight_sums = jnp.sum(weights, axis=0)
    assert_allclose(weight_sums, 1.0, rtol=1e-12, atol=1e-12, err_msg=f'{region_name} weights should sum to 1.0')

    # Adaptive precision testing - determine maximum achievable precision for each nside
    # High nside values face increasing numerical precision challenges
    if nside <= 512:
        precision_threshold = 1e-25
    elif nside <= 1024:
        precision_threshold = 1e-10
    elif nside <= 2048:
        precision_threshold = 1e-8
    elif nside <= 4096:
        precision_threshold = 1e-6
    else:  # nside >= 8192
        precision_threshold = 1e-5

    assert weight_error < precision_threshold, f"""
        {region_name} region nside {nside}: weight MSE {weight_error:.2e}
        exceeds maximum precision threshold {precision_threshold:.2e}
        """

    # Verify all weights are non-negative (physical requirement)
    assert jnp.all(weights >= 0), f'{region_name} all weights should be non-negative'

    # Pixel accuracy validation - be as strict as possible while accounting for high nside complexity
    max_pixel_diff = jnp.max(jnp.abs(sorted_pixels - sorted_hp_pixels))

    # For high nside values, allow minimal pixel differences due to floating point precision limits
    # But still maintain strict bounds relative to the total number of pixels
    if nside <= 1024:
        max_allowed_diff = 0  # Demand perfect pixel matching for smaller nside
    else:
        # For very high nside (2048+), allow minimal differences but keep them extremely small
        # This is much stricter than the original npix//4 tolerance
        max_allowed_diff = min(4, npix // 1000000)  # At most 4 pixels difference, or 1 in a million pixels

    assert max_pixel_diff <= max_allowed_diff, f"""
        {region_name} region nside {nside}: max pixel diff {max_pixel_diff} exceeds strict threshold {max_allowed_diff}
        """


@pytest.mark.parametrize('nside', [4, 8, 16, 32, 64, 128, 256, 512, 8192])
def test_get_interp_weights_shapes(nside):
    """Test that output shapes are correct."""
    # Test single point
    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    pixels, weights = jhp.get_interp_weights(nside, theta, phi)

    # Check shapes
    assert pixels.shape == (4, 1), f'Expected pixels shape (4, 1), got {pixels.shape}'
    assert weights.shape == (4, 1), f'Expected weights shape (4, 1), got {weights.shape}'

    # Test multiple points
    theta = jnp.array([jnp.pi / 4, jnp.pi / 2, 3 * jnp.pi / 4])
    phi = jnp.array([0.0, jnp.pi / 2, jnp.pi])

    pixels, weights = jhp.get_interp_weights(nside, theta, phi)

    # Check shapes
    assert pixels.shape == (4, 3), f'Expected pixels shape (4, 3), got {pixels.shape}'
    assert weights.shape == (4, 3), f'Expected weights shape (4, 3), got {weights.shape}'


@pytest.mark.parametrize('nside', [4, 8, 16, 32, 64, 128, 256, 512, 8192])
def test_get_interp_weights_sum_to_one(nside):
    """Test that weights sum to 1.0 for each point."""
    # Generate test points
    n_points = 10
    theta = jnp.linspace(0.1, jnp.pi - 0.1, n_points)
    phi = jnp.linspace(0.0, 2 * jnp.pi - 0.1, n_points)

    pixels, weights = jhp.get_interp_weights(nside, theta, phi)

    # Check that weights sum to 1 for each point
    weight_sums = jnp.sum(weights, axis=0)
    assert_allclose(weight_sums, 1.0, rtol=1e-12, atol=1e-12)

    # Check that weights are non-negative
    assert jnp.all(weights >= 0), 'All weights should be non-negative'


def test_get_interp_weights_jit():
    """Test that JIT compilation works."""
    nside = 16
    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    # Test that the function can be JIT compiled
    jit_get_interp_weights = jax.jit(jhp.get_interp_weights, static_argnames=['nside'])

    # Should not raise any errors
    pixels, weights = jit_get_interp_weights(nside, theta, phi)

    # Check basic properties
    assert pixels.shape == (4, 1)
    assert weights.shape == (4, 1)
    assert_allclose(jnp.sum(weights), 1.0, rtol=1e-12)


def test_get_interp_weights_gradient():
    """Test that gradient computation works."""
    nside = 16

    def test_func(theta, phi):
        pixels, weights = jhp.get_interp_weights(nside, theta, phi)
        # Return a scalar for gradient testing - sum of weights (should be 1)
        return jnp.sum(weights)

    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    # Should not raise any errors
    grad_func = jax.grad(test_func, argnums=(0, 1))
    grad_theta, grad_phi = grad_func(theta, phi)

    # Check that gradients have the right shape
    assert grad_theta.shape == theta.shape
    assert grad_phi.shape == phi.shape

    # Since sum of weights is always 1, gradients should be 0
    assert_allclose(grad_theta, 0.0, atol=1e-10)
    assert_allclose(grad_phi, 0.0, atol=1e-10)


# Tests for get_interp_val
@pytest.mark.parametrize('nside', [4, 8, 16, 32, 64, 128, 256, 512])
def test_get_interp_val_single_map(nside):
    """Test get_interp_val with single map against healpy."""
    # Create a simple test map
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    # Test single point
    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    jax_result = jhp.get_interp_val(m, theta, phi)
    hp_result = hp.get_interp_val(np.array(m), np.array(theta), np.array(phi))

    assert_allclose(jax_result, hp_result, rtol=1e-12, atol=1e-12)

    # Test multiple points
    theta = jnp.array([jnp.pi / 4, jnp.pi / 2, 3 * jnp.pi / 4])
    phi = jnp.array([0.0, jnp.pi / 2, jnp.pi])

    jax_result = jhp.get_interp_val(m, theta, phi)
    hp_result = hp.get_interp_val(np.array(m), np.array(theta), np.array(phi))

    assert_allclose(jax_result, hp_result, rtol=1e-12, atol=1e-12)


@pytest.mark.parametrize('nside', [4, 8, 16, 32, 64, 128, 256, 512])
def test_get_interp_val_multiple_maps(nside):
    """Test get_interp_val with multiple maps."""
    npix = jhp.nside2npix(nside)

    # Create multiple test maps
    map1 = jnp.arange(npix, dtype=jnp.float64)
    map2 = 2 * jnp.arange(npix, dtype=jnp.float64)
    map3 = jnp.sin(jnp.arange(npix, dtype=jnp.float64))
    maps = jnp.array([map1, map2, map3])

    # Test single point
    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    jax_result = jhp.get_interp_val(maps, theta, phi)

    # Compare with individual map results
    result1 = jhp.get_interp_val(map1, theta, phi)
    result2 = jhp.get_interp_val(map2, theta, phi)
    result3 = jhp.get_interp_val(map3, theta, phi)
    expected = jnp.array([[result1[0]], [result2[0]], [result3[0]]])

    assert_allclose(jax_result, expected, rtol=1e-12, atol=1e-12)

    # Test multiple points
    theta = jnp.array([jnp.pi / 4, jnp.pi / 2])
    phi = jnp.array([0.0, jnp.pi / 2])

    jax_result = jhp.get_interp_val(maps, theta, phi)

    # Should have shape (3, 2) for 3 maps and 2 points
    assert jax_result.shape == (3, 2)


def test_get_interp_val_lonlat():
    """Test get_interp_val with lonlat parameter."""
    nside = 16
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    # Test coordinates in radians
    theta_rad = jnp.array([jnp.pi / 2])
    phi_rad = jnp.array([0.0])

    # Test coordinates in degrees
    lon_deg = jnp.array([0.0])  # 0 degrees longitude
    lat_deg = jnp.array([0.0])  # 0 degrees latitude (equator)

    result_rad = jhp.get_interp_val(m, theta_rad, phi_rad, lonlat=False)
    result_deg = jhp.get_interp_val(m, lon_deg, lat_deg, lonlat=True)

    # Should be the same point
    assert_allclose(result_rad, result_deg, rtol=1e-12, atol=1e-12)


def test_get_interp_val_pixel_mode():
    """Test get_interp_val with theta as pixel indices (phi=None)."""
    nside = 16
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    # Test single pixel
    ipix = jnp.array([100])
    result = jhp.get_interp_val(m, ipix, phi=None)

    # Should be exactly the map value at that pixel
    expected = m[100]
    assert_allclose(result, expected, rtol=1e-12, atol=1e-12)

    # Test multiple pixels
    ipix = jnp.array([50, 100, 150])
    result = jhp.get_interp_val(m, ipix, phi=None)
    expected = m[ipix]
    assert_allclose(result, expected, rtol=1e-12, atol=1e-12)


def test_get_interp_val_against_healpy_comprehensive():
    """Comprehensive test against healpy with various inputs."""
    nside = 16
    npix = jhp.nside2npix(nside)

    # Create a more interesting map (not just linear)
    theta_pix, phi_pix = jhp.pix2ang(nside, jnp.arange(npix))
    m = jnp.sin(2 * theta_pix) * jnp.cos(3 * phi_pix)

    # Test random coordinates
    np.random.seed(42)
    n_test = 20
    theta_test = np.random.uniform(0.1, np.pi - 0.1, n_test)
    phi_test = np.random.uniform(0, 2 * np.pi, n_test)

    jax_result = jhp.get_interp_val(m, theta_test, phi_test)
    hp_result = hp.get_interp_val(np.array(m), theta_test, phi_test)

    # Should match healpy very closely
    assert_allclose(jax_result, hp_result, rtol=1e-10, atol=1e-10)


def test_get_interp_val_shapes():
    """Test that output shapes are correct."""
    nside = 8
    npix = jhp.nside2npix(nside)

    # Single map, single point
    m = jnp.arange(npix, dtype=jnp.float64)
    result = jhp.get_interp_val(m, jnp.array([jnp.pi / 2]), jnp.array([0.0]))
    assert result.shape == (1,), f'Expected shape (1,), got {result.shape}'

    # Single map, multiple points
    theta = jnp.array([jnp.pi / 4, jnp.pi / 2, 3 * jnp.pi / 4])
    phi = jnp.array([0.0, jnp.pi / 2, jnp.pi])
    result = jhp.get_interp_val(m, theta, phi)
    assert result.shape == (3,), f'Expected shape (3,), got {result.shape}'

    # Multiple maps, single point
    maps = jnp.array([m, 2 * m, 3 * m])
    result = jhp.get_interp_val(maps, jnp.array([jnp.pi / 2]), jnp.array([0.0]))
    assert result.shape == (3, 1), f'Expected shape (3, 1), got {result.shape}'

    # Multiple maps, multiple points
    result = jhp.get_interp_val(maps, theta, phi)
    assert result.shape == (3, 3), f'Expected shape (3, 3), got {result.shape}'


def test_get_interp_val_jit():
    """Test that get_interp_val works with JAX JIT compilation."""
    nside = 8
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    # Create JIT compiled version
    jit_get_interp_val = jax.jit(jhp.get_interp_val, static_argnames=['nest', 'lonlat'])

    theta = jnp.array([jnp.pi / 2])
    phi = jnp.array([0.0])

    # Should not raise any errors
    result = jit_get_interp_val(m, theta, phi)

    # Should match non-JIT version
    expected = jhp.get_interp_val(m, theta, phi)
    assert_allclose(result, expected, rtol=1e-12, atol=1e-12)


def test_get_interp_val_gradient():
    """Test that gradients work with get_interp_val."""
    nside = 8
    npix = jhp.nside2npix(nside)

    # Create a smooth map for meaningful gradients
    theta_pix, phi_pix = jhp.pix2ang(nside, jnp.arange(npix))
    m = jnp.sin(theta_pix) * jnp.cos(phi_pix)

    def test_func(theta, phi):
        return jnp.sum(jhp.get_interp_val(m, theta, phi))

    theta = jnp.array([jnp.pi / 3])
    phi = jnp.array([jnp.pi / 4])

    # Should not raise any errors
    grad_func = jax.grad(test_func, argnums=(0, 1))
    grad_theta, grad_phi = grad_func(theta, phi)

    # Check that gradients have the right shape
    assert grad_theta.shape == theta.shape
    assert grad_phi.shape == phi.shape

    # Gradients should be finite (not NaN or inf)
    assert jnp.isfinite(grad_theta).all()
    assert jnp.isfinite(grad_phi).all()


def test_get_interp_val_nest_error():
    """Test that get_interp_val raises error for NEST ordering."""
    nside = 8
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    with pytest.raises(ValueError, match='NEST pixel ordering is not supported'):
        jhp.get_interp_val(m, jnp.array([jnp.pi / 2]), jnp.array([0.0]), nest=True)


def test_get_interp_val_consistency_with_weights():
    """Test that get_interp_val produces the same result as manual calculation with get_interp_weights."""
    nside = 16
    npix = jhp.nside2npix(nside)
    m = jnp.arange(npix, dtype=jnp.float64)

    theta = jnp.array([jnp.pi / 3])
    phi = jnp.array([jnp.pi / 6])

    # Get result from get_interp_val
    result_direct = jhp.get_interp_val(m, theta, phi)

    # Get result by manual calculation
    pixels, weights = jhp.get_interp_weights(nside, theta, phi)
    result_manual = jnp.sum(weights * m[pixels], axis=0)

    # Should be identical
    assert_allclose(result_direct, result_manual, rtol=1e-15, atol=1e-15)



# This file is part of jax-healpy.
# Copyright (C) 2024 CNRS / SciPol developers
#
# jax-healpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jax-healpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jax-healpy. If not, see <https://www.gnu.org/licenses/>.



import chex
import healpy as hp
import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax.errors import TracerBoolConversionError
from jaxtyping import Array
from numpy.testing import assert_array_equal

import jax_healpy as jhp


@pytest.fixture(scope='module', params=['FULL_MAP', 'GAL020', 'GAL040', 'GAL060'])
def mask(request: pytest.FixtureRequest, data_path: str, nside: int) -> tuple[str, Array]:
    if request.param == 'FULL_MAP':
        return request.param, jnp.ones(jhp.nside2npix(nside))
    else:
        return request.param, np.load(f'{data_path}/GAL_PlanckMasks_64.npz')[request.param]


@pytest.fixture(scope='module', params=[64])
def nside(
    request: pytest.FixtureRequest,
) -> int:
    return request.param


def test_kmeans(mask: tuple[str, Array]) -> None:
    name, mask = mask

    (indices,) = jnp.where(mask == 1)

    n_regions = 10
    key = jax.random.key(0)

    clustered = jhp.get_clusters(mask, indices, n_regions, key)
    print(f'Got {n_regions} regions for mask {name}')

    cutout = jhp.get_cutout_from_mask(clustered, indices)
    # Shape must be the same as the mask
    assert cutout.shape == indices.shape

    labels, counts = jnp.unique(cutout, return_counts=True)

    # Check that all the regions are present
    assert_array_equal(labels, jnp.arange(n_regions))

    # Check that number of pixels in each region is close
    assert (counts.std() / counts.mean()) < 0.5

    print(f'all good for mask {name}')


def test_kmeans_jit(mask: tuple[str, Array]) -> None:
    name, mask = mask

    (indices,) = jnp.where(mask == 1)

    n_regions = 10
    key = jax.random.key(0)

    # number of regions cannot be a tracer if max_centroids is None
    jitted_clusters = jax.jit(jhp.get_clusters, static_argnums=(4, 5))
    with pytest.raises(TracerBoolConversionError):
        jitted_clusters(mask, indices, n_regions, key, max_centroids=None)

    #
    # If max_centroids is not None, n_regions can be a tracer and it is jitted once
    @jax.jit
    @chex.assert_max_traces(n=1)
    def jit_clusters(
        mask: Array,
        indices: Array,
        n_regions: Array,
    ) -> Array:
        return jhp.get_clusters(mask, indices, n_regions, key, max_centroids=10)

    _ = jit_clusters(mask, indices, 5)
    _ = jit_clusters(mask, indices, 10)

    chex.clear_trace_counter()

    # If requested number of regions is greater than max_centroids, raise a runtime error

    with pytest.raises(RuntimeError):
        jit_clusters(mask, indices, 20)


def test_cutout_and_reconstruct(mask: tuple[str, Array], nside: int) -> None:
    name, mask = mask

    (indices,) = jnp.where(mask == 1)
    (inv_indices,) = jnp.where(mask != 1)

    gaussian_map = jax.random.normal(jax.random.key(0), jhp.nside2npix(nside))
    # set to unseen everything outside the mask
    gaussian_map = gaussian_map.at[inv_indices].set(jhp.UNSEEN)

    cutout = jhp.get_cutout_from_mask(gaussian_map, indices)

    assert cutout.shape == indices.shape

    reconstruct = jhp.from_cutout_to_fullmap(cutout, indices, nside)

    assert_array_equal(reconstruct, gaussian_map)


def test_frequency_map_cutout(mask: tuple[str, Array], nside: int) -> None:
    # This is usually done to get a cutout out of d the Frequency landscape object from furax

    name, mask = mask

    (indices,) = jnp.where(mask == 1)
    (inv_indices,) = jnp.where(mask != 1)

    frequency_maps = jax.random.normal(jax.random.key(0), (10, jhp.nside2npix(nside)))
    # set to unseen everything outside the mask
    frequency_maps = frequency_maps.at[..., inv_indices].set(jhp.UNSEEN)

    cutout = jhp.get_cutout_from_mask(frequency_maps, indices, axis=1)

    assert cutout.shape == (10, *indices.shape)


def test_normalize_from_clusters(mask: tuple[str, Array], nside: int) -> None:
    name, mask = mask
    (indices,) = jnp.where(mask == 1)

    n_regions = 25
    max_centroids = 50
    key = jax.random.PRNGKey(0)

    # Get raw clusters
    raw_labels = jhp.get_clusters(mask, indices, n_regions, key, max_centroids=max_centroids)
    normalized = jhp.normalize_by_first_occurrence(raw_labels, n_regions, max_centroids)

    # Ensure UNSEEN positions are unchanged
    raw_unseen_mask = raw_labels == hp.UNSEEN
    normalized_unseen_mask = normalized == hp.UNSEEN
    assert_array_equal(raw_unseen_mask, normalized_unseen_mask), 'UNSEEN positions were modified'

    # Remove UNSEEN and validate cluster properties
    valid = np.array(normalized[~normalized_unseen_mask]).astype(np.int64)

    uniques, idx = np.unique(valid, return_index=True)
    assert len(uniques) == n_regions, f'Expected {n_regions} regions, got {len(uniques)}'
    assert (np.sort(idx) == idx).all(), 'First-occurrence indices are not sorted'


def test_shuffle_labels_randomizes_labels():
    arr = np.array([0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 6, 6, 6, 6])
    shuffled = jhp.shuffle_labels(arr)

    # Should contain same elements (set equality), but likely different order
    assert set(shuffled) == {0, 1, 2, 3, 4, 5, 6}
    assert not np.array_equal(arr, shuffled)  # Not equal in order most times


def test_shuffle_labels_preserves_unseen():
    arr = np.array([0, 1, 2, hp.UNSEEN, 1, hp.UNSEEN])
    shuffled = jhp.shuffle_labels(arr)

    # UNSEEN stays where it is
    assert shuffled[3] == hp.UNSEEN
    assert shuffled[5] == hp.UNSEEN

    # All other entries are in the valid label range
    valid_labels = shuffled[shuffled != hp.UNSEEN]
    assert set(valid_labels).issubset({0, 1, 2})



# This file is part of jax-healpy.
# Copyright (C) 2024 CNRS / SciPol developers
#
# jax-healpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jax-healpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jax-healpy. If not, see <https://www.gnu.org/licenses/>.

import sys
from typing import NamedTuple, Optional

import numpy as np
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jax.numpy import arccos, arctan2, cos, deg2rad, newaxis, pi, rad2deg, sin, sqrt
from jaxtyping import Array, PRNGKeyArray

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


_TOL_DEF = 1.0e-5
_MAXITER_DEF = 100
_PIOVER2 = pi * 0.5


class KMeansState(NamedTuple):
    """Holds the state of the KMeans clustering process.

    Attributes:
        ra_dec (Array): The array of RA and DEC coordinates.
        centroids (Array): The array of current centroid coordinates.
        labels (Array): Labels assigning each point to a centroid.
        mean_distance (Array): Current mean distance between points and centroids.
        previous_mean_distance (Array): Previous mean distance for convergence check.
        count (int): Iteration count.
    """

    ra_dec: Array
    centroids: Array
    labels: Array
    mean_distance: Array
    previous_mean_distance: Array
    count: int


class KMeans:
    """KMeans clustering for spherical coordinates using JAX."""

    def __init__(
        self: Self,
        ncenters: int,
        max_centroids: Optional[int] = None,
        tol: float = _TOL_DEF,
        maxiter: int = _MAXITER_DEF,
        initial_sample_size: int = 3,
    ) -> None:
        """Initialize the KMeans instance.

        Args:
            ncenters (int): Number of clusters.
            max_centroids (Optional[int]): Maximum number of centroids to consider.
            tol (float): Tolerance for convergence.
            maxiter (int): Maximum number of iterations.
            initial_sample_size (int): Initial sample size.
        """
        self.ncenters = ncenters
        self.max_centroids = max_centroids
        self.tol = tol
        self.maxiter = maxiter
        self.initial_sample_size = initial_sample_size

    def sample_initial(self: Self, ra_dec: Array, key: PRNGKeyArray) -> tuple[Array, Array]:
        """Sample initial data points and centroids.

        Args:
            ra_dec (Array): Array of RA and DEC coordinates.
            key (PPRNGKeyArrayRNGKey): JAX random key.

        Returns:
            tuple[Array, Array]: Sampled RA/DEC points and initial centroids.
        """
        if self.max_centroids is None:
            nsamples = int(max(2 * np.sqrt(ra_dec.shape[0]), self.initial_sample_size * self.ncenters))
        else:
            nsamples = int(max(2 * np.sqrt(ra_dec.shape[0]), self.initial_sample_size * self.max_centroids))

        sample_key, center_key = jr.split(key, 2)
        if nsamples > ra_dec.shape[0]:
            raise ValueError(
                'Requested centers size is too large for the number of samples. '
                'Consider increasing the nside or decreasing the number of centers (or max_centroids)'
            )

        ra_dec_samples = random_sample(sample_key, ra_dec, nsamples)
        if self.max_centroids is None:
            centroids_samples = random_sample(center_key, ra_dec, self.ncenters)
        else:
            centroids_samples = random_sample(center_key, ra_dec, self.max_centroids)

        return ra_dec_samples, centroids_samples

    def kmeans_init(self: Self, ra_dec: Array, centroids: Array) -> KMeansState:
        """Initialize the KMeans state.

        Args:
            ra_dec (Array): Array of RA and DEC coordinates.
            centroids (Array): Initial centroid coordinates.

        Returns:
            KMeansState: Initialized KMeans state.
        """
        return KMeansState(
            ra_dec=ra_dec,
            centroids=centroids,
            labels=jnp.zeros(ra_dec.shape[0], dtype=jnp.int64),
            mean_distance=jnp.inf,
            previous_mean_distance=0.0,
            count=0,
        )

    def fit(self: Self, ra_dec: Array, centroids: Array) -> KMeansState:
        """Run the KMeans clustering algorithm.

        Args:
            ra_dec (Array): Array of RA and DEC coordinates.
            centroids (Array): Initial centroid coordinates.

        Returns:
            KMeansState: Final state after clustering.
        """
        centroid_mask = jnp.arange(centroids.shape[0]) < self.ncenters

        def kmeans_step(
            carry: tuple[Array, Array, tuple[Array, Array, Array], KMeansState],
        ) -> tuple[Array, Array, tuple[Array, Array, Array], KMeansState]:
            """Perform a single step of the KMeans algorithm.

            Args:
                carry (tuple): Tuple containing current data and state.

            Returns:
                tuple: Updated data and state.
            """
            ra_dec, indices, XYZ, state = carry
            Xs, Ys, Zs = XYZ

            state = state._replace(previous_mean_distance=state.mean_distance)

            distances = cdist_radec(ra_dec, state.centroids)
            if self.max_centroids is not None:
                distances = jnp.where(centroid_mask[None, :], distances, jnp.inf)
            labels = distances.argmin(axis=1).astype(jnp.int64)

            distances = distances[indices, labels]
            mean_distance = distances.mean()

            def for_loop_body(center_indx: int, carry: tuple[Array, Array]) -> tuple[Array, Array]:
                """Update centroids in the for loop.

                Args:
                    center_indx (int): Index of the centroid.
                    carry (tuple): Current centroids and labels.

                Returns:
                    tuple: Updated centroids and labels.
                """
                centroids, labels = carry
                mask = jnp.where(labels == center_indx, 1, 0)
                masked_X = mask * Xs
                masked_Y = mask * Ys
                masked_Z = mask * Zs
                mean_X = masked_X.sum() / mask.sum()
                mean_Y = masked_Y.sum() / mask.sum()
                mean_Z = masked_Z.sum() / mask.sum()

                current_centroid = centroids[center_indx]
                cdistance = xyz2radec(mean_X, mean_Y, mean_Z)
                new_centroid = jnp.where(jnp.isfinite(cdistance), cdistance, current_centroid)

                return centroids.at[center_indx].set(new_centroid), labels

            new_centroids, _ = lax.fori_loop(0, self.ncenters, for_loop_body, (state.centroids, labels))
            new_state = KMeansState(
                ra_dec=ra_dec,
                centroids=new_centroids,
                labels=labels,
                mean_distance=mean_distance,
                previous_mean_distance=state.previous_mean_distance,
                count=state.count + 1,
            )

            return ra_dec, indices, XYZ, new_state

        def kmeans_continue_criteria(carry: tuple[Array, Array, tuple[Array, Array, Array], KMeansState]) -> Array:
            """Check convergence criteria for KMeans.

            Args:
                carry (tuple): Tuple containing current data and state.

            Returns:
                Array: Boolean array indicating whether to continue.
            """
            _, _, _, state = carry

            converged = ((1 - self.tol) * state.previous_mean_distance <= state.mean_distance) & (
                state.previous_mean_distance >= state.mean_distance
            )
            return (state.count < self.maxiter) & (~converged)

        XYZ = radec2xyz(ra_dec[:, 0], ra_dec[:, 1])
        indices = jnp.arange(ra_dec.shape[0])

        init_state = self.kmeans_init(ra_dec, centroids)

        _, _, _, final_state = lax.while_loop(kmeans_continue_criteria, kmeans_step, (ra_dec, indices, XYZ, init_state))

        return final_state


def cdist_radec(a1: Array, a2: Array) -> Array:
    """Compute pairwise spherical distances between two sets of points.

    Args:
        a1 (Array): First array of RA and DEC coordinates.
        a2 (Array): Second array of RA and DEC coordinates.

    Returns:
        Array: Pairwise spherical distances.
    """
    ra1 = a1[:, 0][:, newaxis]
    dec1 = a1[:, 1][:, newaxis]
    ra2 = a2[:, 0]
    dec2 = a2[:, 1]

    phi1, theta1 = deg2rad(ra1), _PIOVER2 - deg2rad(dec1)
    phi2, theta2 = deg2rad(ra2), _PIOVER2 - deg2rad(dec2)

    sintheta1, sintheta2 = sin(theta1), sin(theta2)
    x1, y1, z1 = sintheta1 * cos(phi1), sintheta1 * sin(phi1), cos(theta1)
    x2, y2, z2 = sintheta2 * cos(phi2), sintheta2 * sin(phi2), cos(theta2)

    costheta = x1 * x2 + y1 * y2 + z1 * z2
    costheta = jnp.clip(costheta, -1.0, 1.0)
    return arccos(costheta)


def random_sample(key: PRNGKeyArray, ra_dec: Array, nsamples: int) -> Array:
    """Randomly sample points from the RA/DEC array.

    Args:
        key (PRNGKeyArray): JAX random key.
        ra_dec (Array): Array of RA and DEC coordinates.
        nsamples (int): Number of samples to draw.

    Returns:
        Array: Randomly sampled RA/DEC points.
    """
    nra_dec = ra_dec.shape[0]
    indices = jr.choice(key, nra_dec, shape=(nsamples,), replace=False)
    return ra_dec[indices]


def xyz2radec(x: Array, y: Array, z: Array) -> Array:
    """Convert Cartesian coordinates to spherical (RA/DEC).

    Args:
        x (Array): X coordinate.
        y (Array): Y coordinate.
        z (Array): Z coordinate.

    Returns:
        Array: Array containing RA and DEC.
    """
    r = sqrt(x**2 + y**2 + z**2)
    theta = arccos(z / r)
    phi = arctan2(y, x)
    ra = rad2deg(phi)
    dec = rad2deg(_PIOVER2 - theta)
    ra = atbound1(ra, 0.0, 360.0)
    return jnp.array([ra, dec])


def radec2xyz(ra: Array, dec: Array) -> tuple[Array, Array, Array]:
    """Convert spherical coordinates (RA/DEC) to Cartesian.

    Args:
        ra (Array): Right Ascension.
        dec (Array): Declination.

    Returns:
        tuple[Array, Array, Array]: Cartesian coordinates (X, Y, Z).
    """
    phi, theta = deg2rad(ra), _PIOVER2 - deg2rad(dec)
    sintheta = sin(theta)
    return sintheta * cos(phi), sintheta * sin(phi), cos(theta)


def atbound1(longitude_in: Array, minval: float, maxval: float) -> Array:
    """Ensure longitude is within specified bounds.

    Args:
        longitude_in (Array): Input longitude values.
        minval (float): Minimum bound.
        maxval (float): Maximum bound.

    Returns:
        Array: Longitude values within specified bounds.
    """
    range_size = maxval - minval
    longitude = (longitude_in - minval) % range_size + minval
    return longitude


def kmeans_sample(
    key: PRNGKeyArray,
    ra_dec: Array,
    ncenters: int,
    max_centroids: Optional[int] = None,
    tol: float = _TOL_DEF,
    maxiter: int = _MAXITER_DEF,
    initial_sample_size: int = 3,
) -> KMeansState:
    """Perform KMeans clustering on RA/DEC data.

    Args:
        key (PRNGKeyArray): JAX random key.
        ra_dec (Array): Array of RA and DEC coordinates.
        ncenters (int): Number of clusters.
        max_centroids (Optional[int]): Maximum number of centroids.
        tol (float): Tolerance for convergence.
        maxiter (int): Maximum number of iterations.
        initial_sample_size (int): Initial sample size.

    Returns:
        KMeansState: Final state after clustering.
    """
    if not jnp.isscalar(ncenters):
        raise ValueError('ncenters must be a scalar')
    km = KMeans(ncenters, max_centroids, tol, maxiter, initial_sample_size=initial_sample_size)
    ra_dec_samples, centroids_samples = km.sample_initial(ra_dec, key)

    state = km.fit(ra_dec_samples, centroids_samples)
    state = km.fit(ra_dec, state.centroids)

    return state



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath('../../'))

project = 'jax-healpy'
copyright = '2024, Pierre Chanial, Simon Biquard, Wassim Kabalan'
author = 'Pierre Chanial, Simon Biquard, Wassim Kabalan'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'myst_parser',
    'nbsphinx',
    'sphinx_copybutton',
]

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']

# Theme options
html_theme_options = {
    'repository_url': 'https://github.com/pchanial/jax-healpy',
    'use_repository_button': True,
    'use_issues_button': True,
    'use_edit_page_button': True,
    'path_to_docs': 'docs/source',
    'use_download_button': True,
    'use_sidenotes': True,
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Intersphinx mapping
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'jax': ('https://jax.readthedocs.io/en/latest/', None),
    'healpy': ('https://healpy.readthedocs.io/en/latest/', None),
}

# MyST parser configuration
myst_enable_extensions = [
    'dollarmath',
    'amsmath',
    'deflist',
    'fieldlist',
    'html_admonition',
    'html_image',
    'colon_fence',
    'smartquotes',
    'replacements',
    'strikethrough',
    'tasklist',
]

# Nbsphinx configuration
nbsphinx_execute = 'never'  # Don't execute notebooks during build

# Autodoc configuration
autodoc_member_order = 'bysource'
autodoc_typehints = 'description'
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'show-inheritance': True,
}

# Copy button configuration
copybutton_prompt_text = r'>>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: '
copybutton_prompt_is_regexp = True



import numpy as np
import pytest
from numpy.testing import assert_allclose

import jax_healpy as hp


@pytest.mark.parametrize(
    'z, iphi',
    [
        (1, 0),
        (0.999, 0),
        (0.999, 1),
        (0.999, 2),
        (0.999, 3),
        (0.999, 4),
        (0.98, 0),
        (0.98, 1),
        (0.98, 2),
        (0.98, 3),
        (0.98, 4),
        (0, 0),
        (0, 1),
        (0, 2),
        (0, 3),
        (0, 4),
        (-0.98, 0),
        (-0.98, 1),
        (-0.98, 2),
        (-0.98, 3),
        (-0.98, 4),
        (-0.999, 0),
        (-0.999, 1),
        (-0.999, 2),
        (-0.999, 3),
        (-0.999, 4),
        (-1, 0),
        (-1, 1),
        (-1, 2),
        (-1, 3),
        (-1, 4),
    ],
)
@pytest.mark.parametrize('lonlat', [False, True])
def test_ang2vec2ang(z: float, iphi: int, lonlat: bool) -> None:
    # test that we get the same results as healpy
    theta = np.arccos(z)
    phi = iphi / 5 * 2 * np.pi
    actual_theta, actual_phi = hp.vec2ang(hp.ang2vec(theta, phi, lonlat=lonlat), lonlat=lonlat)
    assert_allclose(actual_theta, theta, rtol=1e-14, atol=1e-15)
    assert_allclose(actual_phi, phi, rtol=1e-14, atol=1e-15)


@pytest.mark.parametrize('lonlat', [False, True])
def test_ang2vec_array(lonlat: bool) -> None:
    theta = np.array([np.pi / 4, 3 * np.pi / 4])
    phi = np.array([np.pi / 2, 3 * np.pi / 2])
    vec0 = hp.ang2vec(theta[0], phi[0], lonlat=lonlat)
    assert vec0.shape == (3,)
    vec1 = hp.ang2vec(theta[1], phi[1], lonlat=lonlat)
    vec = hp.ang2vec(theta, phi, lonlat=lonlat)
    assert vec.shape == (2, 3)
    assert_allclose(vec, np.array([vec0, vec1]), rtol=1e-14)


@pytest.mark.parametrize('lonlat', [False, True])
def test_vec2ang_array(lonlat: bool) -> None:
    vec = np.array([[1, 2, 3], [-1, 2, -1]])
    theta0, phi0 = hp.vec2ang(vec[0], lonlat=lonlat)
    assert theta0.shape == (1,)
    assert phi0.shape == (1,)
    theta1, phi1 = hp.vec2ang(vec[1], lonlat=lonlat)
    theta, phi = hp.vec2ang(vec, lonlat=lonlat)
    assert theta.shape == (2,)
    assert phi.shape == (2,)
    assert_allclose(theta, np.array([theta0[0], theta1[0]]), rtol=1e-14)
    assert_allclose(phi, np.array([phi0[0], phi1[0]]), rtol=1e-14)



from typing import Any, Callable

import numpy as np
import pytest

s2fft = pytest.importorskip('s2fft')

from s2fft.sampling.s2_samples import flm_2d_to_hp  # noqa: E402
from s2fft.utils import signal_generator  # noqa: E402


@pytest.fixture
def flm_generator(numpy_rng) -> Callable[[...], np.ndarray]:
    # Import s2fft (and indirectly numpy) locally to avoid
    # `RuntimeWarning: numpy.ndarray size changed` when importing at module level
    # import s2fft as s2f
    # from s2fft.utils import signal_generator
    def generate_flm(L: int, healpy_ordering: bool = False, **keywords: Any) -> np.ndarray:
        flm = signal_generator.generate_flm(numpy_rng, L, **keywords)
        if healpy_ordering:
            flm = flm_2d_to_hp(flm, L)
        return flm

    return generate_flm



import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from numpy.testing import assert_array_equal

import jax_healpy as hp
from jax_healpy.pixelfunc import MAX_NSIDE


@pytest.mark.parametrize(
    'nside, x, y, expected_fpix',
    [
        (1, 0, 0, 0),
        (2, 0, 0, 0),
        (2, 1, 1, 3),
        (4, 1, 2, 9),
        (4, 0, 3, 10),
        (4, 3, 0, 5),
        (MAX_NSIDE, MAX_NSIDE - 1, MAX_NSIDE - 1, MAX_NSIDE**2 - 1),
    ],
)
def test_xy_to_fpix(nside, x, y, expected_fpix):
    fpix = hp.pixelfunc._xy2fpix(nside, x, y)
    assert_array_equal(fpix, expected_fpix)


@pytest.mark.parametrize(
    'nside, fpix, expected_x, expected_y',
    [
        (1, 0, 0, 0),
        (2, 0, 0, 0),
        (2, 3, 1, 1),
        (4, 9, 1, 2),
        (4, 10, 0, 3),
        (4, 5, 3, 0),
        (MAX_NSIDE, MAX_NSIDE**2 - 1, MAX_NSIDE - 1, MAX_NSIDE - 1),
    ],
)
def test_fpix_to_xy(nside, fpix, expected_x, expected_y):
    x, y = hp.pixelfunc._fpix2xy(nside, fpix)
    assert_array_equal(x, expected_x)
    assert_array_equal(y, expected_y)


@pytest.mark.parametrize('order', range(30))
@pytest.mark.parametrize('nest', [True, False])
def test_pix_to_xyf_to_pix(order: int, nest: bool) -> None:
    nside = hp.order2nside(order)
    npix = hp.nside2npix(nside)
    maxpix = 1_000
    if npix <= maxpix:
        # up to nside=4, test all pixels
        pix = jnp.arange(npix)
    else:
        # otherwise only test a random subset
        # not using jr.choice(replace=False) because it would be slow for large npix
        # it is not a problem if some pixels are tested twice
        pix = jr.randint(jr.key(1234), shape=(maxpix,), minval=0, maxval=npix)
    xyf = hp.pix2xyf(nside, pix, nest=nest)
    ppix = hp.xyf2pix(nside, *xyf, nest=nest)
    assert_array_equal(ppix, pix)


@pytest.mark.parametrize(
    'nside, x, y, f, expected_pix',
    [
        (16, 8, 8, 4, 1440),
        (16, [8, 8, 8, 15, 0], [8, 8, 7, 15, 0], [4, 0, 5, 0, 8], [1440, 427, 1520, 0, 3068]),
    ],
)
def test_xyf2pix(nside: int, x: int, y: int, f: int, expected_pix: int):
    assert_array_equal(hp.xyf2pix(nside, x, y, f), expected_pix)


@pytest.mark.parametrize(
    'nside, ipix, expected_xyf',
    [
        (16, 1440, ([8], [8], [4])),
        (16, [1440, 427, 1520, 0, 3068], ([8, 8, 8, 15, 0], [8, 8, 7, 15, 0], [4, 0, 5, 0, 8])),
        pytest.param(
            (1, 2, 4, 8),
            11,
            ([0, 1, 3, 7], [0, 0, 2, 6], [11, 3, 3, 3]),
            marks=pytest.mark.xfail(reason='nside must be an int'),
        ),
    ],
)
def test_pix2xyf(nside, ipix, expected_xyf):
    x, y, f = hp.pix2xyf(nside, ipix)
    assert_array_equal(x, expected_xyf[0])
    assert_array_equal(y, expected_xyf[1])
    assert_array_equal(f, expected_xyf[2])


@pytest.mark.parametrize(
    'nside, ipix_ring, expected_ipix_nest',
    [
        (16, 1504, 1130),
        (2, np.arange(10), [3, 7, 11, 15, 2, 1, 6, 5, 10, 9]),
        pytest.param(
            [1, 2, 4, 8],
            11,
            [11, 13, 61, 253],
            marks=pytest.mark.xfail(reason='nside must be an int'),
        ),
    ],
)
def test_ring2nest(nside: int, ipix_ring: int, expected_ipix_nest: int):
    assert_array_equal(hp.ring2nest(nside, ipix_ring), expected_ipix_nest)


@pytest.mark.parametrize(
    'nside, ipix_nest, expected_ipix_ring',
    [
        (16, 1130, 1504),
        (2, np.arange(10), [13, 5, 4, 0, 15, 7, 6, 1, 17, 9]),
        pytest.param(
            [1, 2, 4, 8],
            11,
            [11, 2, 12, 211],
            marks=pytest.mark.xfail(reason='nside must be an int'),
        ),
    ],
)
def test_nest2ring(nside: int, ipix_nest: int, expected_ipix_ring: int):
    assert_array_equal(hp.nest2ring(nside, ipix_nest), expected_ipix_ring)


@pytest.mark.parametrize('func_name', ['ring2nest', 'nest2ring'])
def test_ring_nest_dtypes(func_name: str):
    small_nside = 512
    large_nside = 16384
    pix32 = jnp.zeros(10, dtype=jnp.int32)
    pix64 = jnp.zeros(10, dtype=jnp.int64)
    func = getattr(hp, func_name)
    assert func(small_nside, pix32).dtype == jnp.int32  # input is int32, so should be output
    assert func(small_nside, pix64).dtype == jnp.int64  # input is int64, so should be output
    assert func(large_nside, pix32).dtype == jnp.int64  # nside is large so output is int64
    assert func(large_nside, pix64).dtype == jnp.int64  # nside is large so output is int64


@pytest.mark.parametrize(
    'order', list(range(29)) + [pytest.param(30, marks=pytest.mark.xfail(reason='overflow somewhere?'))]
)
def test_roundtrip_max_pixel(order):
    nside = hp.order2nside(order)
    npix = hp.nside2npix(nside)
    max_pix = npix - 1
    assert hp.ring2nest(nside, hp.nest2ring(nside, max_pix)) == max_pix
    assert hp.nest2ring(nside, hp.ring2nest(nside, max_pix)) == max_pix


@pytest.mark.parametrize('r2n, n2r', [(True, False), (False, True)])
def test_reorder_base_pixels(r2n, n2r):
    map_in = jnp.arange(12)
    map_out = hp.reorder(map_in, r2n=r2n, n2r=n2r)
    assert_array_equal(map_out, map_in)


def test_r2n():
    map_in = jnp.arange(48)
    map_out = hp.reorder(map_in, r2n=True)
    assert_array_equal(map_out, [13,  5,  4,  0, 15,  7,  6,  1, 17,  9,  8,  2, 19, 11, 10,  3, 28,
                                 20, 27, 12, 30, 22, 21, 14, 32, 24, 23, 16, 34, 26, 25, 18, 44, 37,
                                 36, 29, 45, 39, 38, 31, 46, 41, 40, 33, 47, 43, 42, 35])  # fmt: skip


def test_n2r():
    map_in = jnp.arange(48)
    map_out = hp.reorder(map_in, n2r=True)
    assert_array_equal(map_out, [ 3,  7, 11, 15,  2,  1,  6,  5, 10,  9, 14, 13, 19,  0, 23,  4, 27,
                                  8, 31, 12, 17, 22, 21, 26, 25, 30, 29, 18, 16, 35, 20, 39, 24, 43,
                                 28, 47, 34, 33, 38, 37, 42, 41, 46, 45, 32, 36, 40, 44])  # fmt: skip


@pytest.mark.parametrize('r2n, n2r', [(True, False), (False, True)])
@pytest.mark.parametrize('shape', [(12,), (3, 12), (2, 3, 12)])
def test_reorder_shape(shape: tuple[int], r2n: bool, n2r: bool):
    map_out = hp.reorder(jnp.zeros(shape), r2n=r2n, n2r=n2r)
    assert map_out.shape == shape


@pytest.mark.parametrize('order', range(8))
@pytest.mark.parametrize('batch', [(), (1,)])
def test_reorder_roundtrip(order: int, batch: tuple[int]):
    npix = hp.order2npix(order)
    map_in = jr.uniform(jr.key(1143), batch + (npix,))
    assert_array_equal(hp.reorder(hp.reorder(map_in, r2n=True), n2r=True), map_in)
